import copy
import time
import logging
import sys
import yaml
from PyQt5 import QtWidgets, QtCore, QtGui
import pydantic
from pydantic.color import Color as pydColor
import typing
from redvypr.device import RedvyprDevice
from redvypr.widgets.redvyprAddressWidget import RedvyprAddressWidgetSimple, RedvyprAddressWidget, datastreamQTreeWidget
import redvypr.files as files
from redvypr.redvypr_address import RedvyprAddress
from redvypr.data_packets import RedvyprMetadata, RedvyprMetadataGeneral


logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('pydanticConfigWidget')
logger.setLevel(logging.DEBUG)


class pydanticDeviceConfigWidget(QtWidgets.QWidget):
    """
    Config widget for a pydantic configuration
    """
    def __init__(self, device=None, exclude=[], config_location='right', show_datatype=False):
        funcname = __name__ + '.__init__():'
        super().__init__()
        logger.debug(funcname)
        self.setWindowTitle('Config of {}'.format(device.name))
        self.device = device
        try:
            redvypr = self.device.redvypr
        except:
            redvypr = None
        self.exclude = exclude
        self.layout = QtWidgets.QGridLayout(self)
        #self.label = QtWidgets.QLabel('Configuration of\n{:s}'.format(self.device.name))
        #self.layout.addWidget(self.label)
        dataname = self.device.name + '.config'
        config = self.device.custom_config
        if config is None:
            logger.warning('No config existing')
            self.configWidget = QtWidgets.QLabel('No config existing!')
        else:
            #print('Config to edit',self.device.custom_config)
            #print('tpye',type(self.device.custom_config))
            self.configWidget = pydanticConfigWidget(self.device.custom_config, configname=dataname, exclude=self.exclude, config_location=config_location, show_datatype=show_datatype, redvypr=redvypr)
            self.configWidget.config_changed_flag.connect(self.config_changed)
            self.configWidget.config_editing_done.connect(self.closeClicked)
            #self.configWidget = pydanticQTreeWidget(self.device.custom_config, dataname=dataname, exclude=self.exclude)
        self.layout.addWidget(self.configWidget)

    def config_changed(self):
        funcname = __name__ + '.config_changed():'
        logger.debug(funcname)
        self.device.config_changed()

    def closeClicked(self):
        funcname = __name__ + '.closeClicked():'
        logger.debug(funcname)
        self.close()

#
#
# pydanticConfigWidget
#
#
class pydanticConfigWidget(QtWidgets.QWidget):
    #config_changed = QtCore.pyqtSignal(dict)  # Signal notifying that the configuration has changed
    config_changed_flag = QtCore.pyqtSignal()  # Signal notifying that the configuration has changed
    config_editing_done = QtCore.pyqtSignal()
    def __init__(self, config=None, editable=True, configname=None, exclude=[], config_location='right', show_datatype=False, redvypr=None, show_editable_only=True, close_after_editing=True):
        funcname = __name__ + '.__init__():'
        super().__init__()
        self.redvypr = redvypr
        self.exclude = exclude
        self.layout = QtWidgets.QGridLayout(self)
        self.config_location = config_location
        if self.config_location == 'bottom':
            sdir = QtCore.Qt.Vertical
        else:
            sdir = QtCore.Qt.Horizontal
        self.splitter = QtWidgets.QSplitter(sdir)
        self.close_after_editing = close_after_editing
        #self.label = QtWidgets.QLabel('Configuration of\n{:s}'.format(self.device.name))
        #self.layout.addWidget(self.label)
        if configname is None:
            configname = 'pydanticConfig'

        self.config = config
        self.configWidget = pydanticQTreeWidget(self.config, dataname=configname, exclude=self.exclude, show_datatype=show_datatype, show_editable_only=show_editable_only)
        if editable:
            self.configWidget.itemDoubleClicked.connect(self.__openConfigGui__)
            self.configGui = QtWidgets.QWidget()  # Widget where the user can modify the content
            self.configGui_layout = QtWidgets.QVBoxLayout(self.configGui)
            self.edit_label = QtWidgets.QLabel('Edit data')
            self.configGui_layout.addWidget(self.edit_label)
            self.stretchy_spacer_thing = QtWidgets.QSpacerItem(10,10,QtWidgets.QSizePolicy.Minimum,QtWidgets.QSizePolicy.Expanding)
            self.configGui_layout.addItem(self.stretchy_spacer_thing)
            # Add a blank widget for editing
            self.__configwidget = QtWidgets.QWidget()
            self.configGui_layout.addWidget(self.__configwidget)
            #if self.config_location == 'bottom':
            #    self.layout.addWidget(self.configGui, 1, 0)
            #else:
            #    self.layout.addWidget(self.configGui, 0, 1)


        self.closeButton = QtWidgets.QPushButton('Apply')
        self.closeButton.clicked.connect(self.closeClicked)

        self.cancelButton = QtWidgets.QPushButton('Cancel')
        self.cancelButton.clicked.connect(self.cancelClicked)

        #self.layout.addWidget(self.configWidget, 0, 0)
        self.splitter.addWidget(self.configWidget)
        self.splitter.addWidget(self.configGui)
        self.layout.addWidget(self.splitter, 0, 0)
        self.layout.addWidget(self.closeButton, 2, 0, 1, 1)
        self.layout.addWidget(self.cancelButton, 2, 1, 1, 1)

    def cancelClicked(self):
        funcname = __name__ + '.cancel():'
        logger.debug(funcname)
        if self.close_after_editing:
            self.close()

    def closeClicked(self):
        funcname = __name__ + '.close():'
        logger.debug(funcname)
        self.config_editing_done.emit()
        if self.close_after_editing:
            self.close()

    def __comboUpdateItem(self, index):
        """
        I called when the combo that allows the user to choose a datatype is changed
        :param index:
        :return:
        """
        funcname = __name__ + '.__comboUpdateItem():'
        logger.debug(funcname)
        user_role_config = 11
        role = QtCore.Qt.UserRole + user_role_config
        item = self.__configCombo.itemData(index, role)
        # Get the datatypestr and call the gui create function again
        datatypestr = self.__configCombo.currentText()
        logger.debug(funcname + 'Datatypestr {}'.format(datatypestr))
        item.__datatypestr__ = datatypestr

        # Get the type_hint object and create standard type
        #dobject = typing.get_args(item.__type_hints__)[index]  # This is not working with unions, using interprete type_hints instead
        type_dict = self.interprete_type_hints(item.__type_hints__)
        dobject = type_dict['datatype_objects'][datatypestr]
        print('type dict',type_dict)
        print('Item', item)
        print('Index',index)
        print('Item type hints', item.__type_hints__)
        print('Item datatypestr', datatypestr)
        print('Dobject',dobject)
        logger.debug(funcname + ' object : {}'.format(dobject))
        try:
            flag_add = item.__flag_add__
        except:
            flag_add = False

        if flag_add:
            logger.debug(funcname + ' Add flag')
        else:
            logger.debug(funcname + ' Modifying item')
            if 'literal' in dobject.__name__.lower():
                logger.debug(funcname + 'Literal type hint')
                literal_args = typing.get_args(dobject)
                # Get the first element
                data = literal_args[0]
                item.__datatypestr__ = 'literal'
                item.__literal_options__ = literal_args
            elif 'color' in dobject.__name__.lower():
                logger.debug(funcname + 'Color type hint')
                item.__data__ = dobject('red')
            else:
                item.__data__ = dobject()

            logger.debug(funcname + 'Index: {} datatpystr: {} data: {}'.format(index,datatypestr, item.__data__))


    def __comboTypeChanged(self,index):
        """
        Called when the combobox of the datatype is changed
        :param index:
        :return:
        """
        funcname = __name__ + '__comboTypeChanged():'
        logger.debug(funcname)
        #print('Index', index)
        self.__comboUpdateItem(index)
        user_role_config = 11
        role = QtCore.Qt.UserRole + user_role_config
        item = self.__configCombo.itemData(index, role)


        self.__configwidget.setParent(None)
        #self.configGui_layout.addWidget(self.__configwidget)
        self.configGui_layout.removeItem(self.stretchy_spacer_thing)

        try:
            flag_add = item.__flag_add__
        except:
            flag_add = False

        if flag_add:
            logger.debug(funcname + ' Add flag')
        else:
            self.CreateConfigWidgetForItem(item)
        #print('Item',item)
        #print('Clear gui')
        #self.__clearConfigGui__()
        #print('Populate gui')
        self.__populateConfigGui__()
        #self.__openConfigGui__(item)

        #print('Open config gui with item',item)
        # Changed the datatype, now recreate the widgets
        #self.configGui_layout.removeItem(self.stretchy_spacer_thing)
        #try:
        #    self.__configwidget.close()
        #except:
        #    pass


        #self.configGui_layout.addWidget(self.__configwidget)
        ## Add a stretch
        #self.configGui_layout.addItem(self.stretchy_spacer_thing)

    def interprete_type_hint_annotation(self, type_hint):
        funcname = __name__ + '.interprete_type_hint_annotation():'
        isannotated = 'Annotated' in type_hint.__class__.__name__
        print(funcname + 'type hint',type_hint)
        if isannotated:
            logger.debug(funcname + 'Disentangling the annotations')
            # If annotated data is available look search for known datatypes
            annotations = typing.get_args(type_hint)
            print('Annotations ...', annotations)
            for annotation in annotations:
                if annotation == 'RedvyprAddressStr':
                    #return [str, 'RedvyprAddressStr']
                    return [RedvyprAddress, 'RedvyprAddressStr']

            type_hint = typing.get_args(type_hint)[0]
            print('type hint new', type_hint)
            return [type_hint, type_hint.__name__]
        else:
            print('type hint new else', type_hint)
            try:
                type_str = type_hint.__name__
            except:
                type_str = 'NA'
            return [type_hint, type_str]


    def interprete_type_hints(self, type_hints):
        funcname = __name__ + '.interprete_type_hints():'
        type_dict = None
        #print('Type hints', type_hints)
        if type_hints is not None:
            #print('Name', type_hints.__name__)
            type_dict = {'type_args':[],'datatype_objects':{}}
            # Get the different type hints

            #print('Type args', len(type_args), type(type_hints))
            index_datatype = None
            # Create a combo to allow the user to choose between the different
            # data type choices
            # checks if type hints are annotations or types/unions
            # Annotations are used for extra information, type/unions for
            # And are ignored
            #print('Classname', type_hints.__class__.__name__)
            isliteral = 'Literal' in type_hints.__class__.__name__
            type_dict['isliteral'] = isliteral
            #isannotated = 'Annotated' in type_hints.__class__.__name__
            #type_dict['isannotated'] = isannotated
            [type_hints, tmp] = self.interprete_type_hint_annotation(type_hints)
            type_args = typing.get_args(type_hints)
            type_dict['type_root'] = None
            if 'union' in type_hints.__name__.lower():
                logger.debug(funcname + ' Found Union of datatypes in type hints')
                type_dict['type_root'] = 'union'
                #type_dict['isannotated'] = False
            #logger.debug(funcname + 'Isannotated {}'.format(isannotated))
            if isliteral:
                type_dict['literal_options'] = type_args
            #elif len(type_args) > 0 and not (isannotated):
            elif len(type_args):
                # Remove str from args, as this is a mandatary entry for dict, but not needed
                if 'dict' in type_hints.__name__.lower() and (type_args[0].__name__.lower() == 'str' or type_args[0].__name__.lower() == 'int'):
                    logger.debug(funcname + 'Removing str/int entry for dict')
                    type_args = type_args[1]
                    [type_args, argstr] = self.interprete_type_hint_annotation(type_args)
                    print('type args',type_args, 'argstr', argstr)
                    if 'union' in argstr.lower():
                        print(funcname + 'Union')
                        type_args = typing.get_args(type_args)

                print('type args',type_args)
                [type_args, tmp] = self.interprete_type_hint_annotation(type_args)
                # loop over all type args, if there is a union within (i.e. a list with some datatypes), loop over the union
                # for example typing.List[typing.Union[float, str]]
                for arg in type_args:
                    logger.debug(funcname + 'arg {}'.format(arg))
                    [arg, argstr] = self.interprete_type_hint_annotation(arg)
                    if 'union' in argstr.lower():
                        logger.debug(funcname + ' Found Union of datatypes')
                        type_dict['type_root'] = 'union'
                        type_args_union = typing.get_args(arg)
                        for arg_union in type_args_union:
                            [arg_union, argstr] = self.interprete_type_hint_annotation(arg_union)
                            logger.debug(funcname + 'Adding datatype: {}'.format(argstr))
                            type_dict['type_args'].append(argstr)
                            # Add the object as well
                            type_dict['datatype_objects'][argstr] = arg_union
                    else:
                        type_dict['type_args'].append(argstr)
                        # Add the object as well
                        type_dict['datatype_objects'][argstr] = arg

        return type_dict

    def __populateConfigGui__(self):
        logger.debug('Additional widgets {}'.format(self.additional_config_gui_widgets))
        for w in self.additional_config_gui_widgets:
            try:
                logger.debug('Adding widget {}'.format(w))
                self.configGui_layout.addWidget(w)
            except:
                logger.debug('Could not add widget',exc_info=True)

        self.configGui_layout.addWidget(self.__configwidget)
        self.configGui_layout.addItem(self.stretchy_spacer_thing)

    def __clearConfigGui__(self):
        self.configGui_layout.removeItem(self.stretchy_spacer_thing)
        for i in reversed(range(self.configGui_layout.count())):
            self.configGui_layout.itemAt(i).widget().setParent(None)
            #self.configGui_layout.itemAt(i).widget().hide()


    def __openConfigGui__(self, item):
        """
        Function is called when an item is double clicked. It creates the
        config widget.
        :param item:
        :return:
        """
        funcname = __name__ + '.__openConfigGui__():'
        logger.debug(funcname + ' for item {}'.format(item.text(0)))
        has_combo = False
        self.additional_config_gui_widgets = []
        self.__clearConfigGui__()
        editable = item.__editable__
        #print('editable',editable)
        flag_add = False
        if editable==False:
            #print('Not editable')
            pass
        else:
            #print('some information')
            #print(item.__data__)
            #print(item.__dataparent__)
            #print(item.__dataindex__)
            #print(item.__datatypestr__)
            #print(item.__parent__)
            #print('Done')
            try:
                type_hints = item.__type_hints__
            except:
                type_hints = None

            type_dict = self.interprete_type_hints(type_hints)
            #print(funcname + ' Type hints', type_hints)
            #print(funcname + ' Type dict', type_dict)
            if type_dict is not None:
                index_datatype = None
                # Create a combo to allow the user to choose between the different
                # data type choices
                # checks if type hints are annotations or types/unions
                # Annotations are used for extra information, type/unions for
                if type_dict['isliteral']:
                    #print('Literal options',type_dict['literal_options'])
                    item.__datatypestr__ = 'literal'
                    item.__datatype__ = typing.Literal
                    item.__literal_options = type_dict['literal_options']
                #elif type_dict['isannotated']:
                #    pass
                # This is a union of allowed types
                elif (len(type_dict['type_args']))>0 and (type_hints.__name__.lower() == 'union'):
                    logger.debug(funcname + 'Type args')
                    self.__configCombo = QtWidgets.QComboBox()
                    has_combo = True
                    logger.debug(funcname + 'Filling combo box')
                    for iarg,argstr in enumerate(type_dict['type_args']):
                        logger.debug('Argstr {}'.format(argstr))
                        self.__configCombo.addItem(argstr)
                        index = self.__configCombo.count() - 1
                        datatypestr_tmp = item.__data__.__class__.__name__
                        if argstr == datatypestr_tmp:
                            index_datatype = index
                        user_role_config = 11
                        role = QtCore.Qt.UserRole + user_role_config
                        self.__configCombo.setItemData(index, item, role)
                        #print('index',index,'item',item,'argstr',argstr, datatypestr_tmp, item.__datatypestr__)

                    data = item.__data__
                    # Check if the combo is meant to choose between different types
                    if type_dict['type_root'] == 'union':
                        logger.debug(funcname + 'Union')
                        # Add the config combo to the layout
                        self.additional_config_gui_widgets.append(self.__configCombo)
                    else: # Ordinary item, adding nothing special
                        logger.debug(funcname + 'Ordinary item')
                        ## Add the config combo to the layout
                        #self.additional_config_gui_widgets.append(self.__configCombo)

                    item_data = item.__data__
                    if index_datatype is not None:
                        self.__configCombo.setCurrentIndex(index_datatype)
                        item.__datatypestr__ = datatypestr_tmp
                        item.__datatype__ = item.__data__.__class__
                    else:
                        self.__configCombo.setCurrentIndex(0)
                        # self.__comboUpdateItem(0)

                    # Get the index for the combo
                    self.__configCombo.currentIndexChanged.connect(self.__comboTypeChanged)

            self.CreateConfigWidgetForItem(item)
            # Do we still need this?
            if flag_add:
                item.__flag_add__ = True
            else:
                item.__flag_add__ = False
            # Add a remove button if the item can be removed
            parentdata = item.__dataparent__
            try:
                removable = item.__removable__
            except:
                removable = False
            if removable:
                self.__remove_button = QtWidgets.QPushButton('Remove')
                self.__remove_button.item = item
                self.__remove_button.clicked.connect(self.__removeClicked)
                self.additional_config_gui_widgets.append(QtWidgets.QLabel('Remove entry'))
                self.additional_config_gui_widgets.append(self.__remove_button)

            self.__populateConfigGui__()
            if has_combo:
                self.__configCombo.setCurrentIndex(0)

    def __removeClicked(self):
        funcname = __name__ + '.__removeClicked():'
        logger.debug(funcname)
        item = self.sender().item
        parentdata = item.__dataparent__
        if isinstance(parentdata, dict):
            #print('Removing from dict')
            parentdata.pop(item.__dataindex__,None)
        elif isinstance(parentdata, list):
            #print('Removing from list')
            parentdata.pop(item.__dataindex__)
        elif pydantic.BaseModel in parentdata.__class__.__mro__:
            #print('Removing from basemodel')
            delattr(parentdata, item.__dataindex__)

        self.__clearConfigGui__()
        # Reload and redraw all data
        self.configWidget.reload_data()
        self.config_changed_flag.emit()
        #print(item.__data__)
        #print(item.__dataparent__)
        #print(item.__dataindex__)
        #print(item.__datatypestr__)
        #print(item.__parent__)

    def CreateConfigWidgetForItem(self, item):
        """
        creates self.__configwidget for item type
        :param item:
        :return:
        """
        funcname = __name__ + '.CreateConfigWidgetForItem():'
        try:
            datatypestr = item.__datatypestr__
        except:
            datatypestr = 'NA'

        logger.debug(funcname + 'Datatypestr {}'.format(datatypestr))
        #print('MRO',item.__data__.__class__.__mro__)

        try:
            self.__configwidget.close()
        except:
            pass

        #if pydantic.BaseModel in item.__data__.__class__.__mro__:
        #    print('Existing Basemodel ...')
        #    self.createConfigWidgetBaseModel(item)
        logger.debug(funcname + 'Item datatypestr {}'.format(item.__datatypestr__))
        if (item.__datatypestr__ == 'int') or (item.__datatypestr__ == 'float'):
            self.createConfigWidgetNumber(item, dtype=item.__datatypestr__)
        elif (item.__datatypestr__.lower() == 'color'):
            logger.debug(funcname + 'Color datatype')
            self.createConfigWidgetColor(item)
        elif (item.__datatypestr__.lower() == 'literal'):
            logger.debug('Literal datatype')
            self.createConfigWidgetLiteral(item)
        elif (item.__datatypestr__ == 'str'):
            logger.debug('Str')
            self.createConfigWidgetStr(item)
        elif (item.__datatypestr__ == 'datetime'):
            self.createConfigWidgetDateTime(item)
        elif (item.__datatypestr__ == 'bool'):
            self.createConfigWidgetBool(item)
        elif (item.__datatypestr__ == 'RedvyprAddress'):
            logger.debug('RedvyprAddress')
            self.createConfigWidgetRedvyprAddressStr(item)
        elif (item.__datatypestr__ == 'RedvyprAddressStr'):
            logger.debug('RedvyprAddressStr')
            self.createConfigWidgetRedvyprAddressStr(item)
        elif (item.__datatypestr__ == 'list'):
            self.createConfigWidgetList(item)
        elif (item.__datatypestr__ == 'dict'):
            logger.debug(funcname + 'Dictionary')
            self.createConfigWidgetDict(item)
        elif (item.__datatypestr__ == 'bytes'):
            logger.debug(funcname + 'bytes')
            self.createConfigWidgetBytes(item)
        elif pydantic.BaseModel in item.__data__.__class__.__mro__:
            logger.debug(funcname + 'BaseModel')
            self.createConfigWidgetBaseModelNew(item)
        else:
            try: # Check if there is a valid type_hint, otherwise do nothing
                type_hints = item.__type_hints__
                type_dict = self.interprete_type_hints(type_hints)
                dobject = type_dict['datatype_objects'][item.__datatypestr__]
                self.createConfigWidgetBaseModelNew(item)
            except:
                logger.debug('Could not create a widget:',exc_info=True)

    def createConfigWidgetColor(self, item):
        funcname = __name__ + '.createConfigWidgetColor():'
        logger.debug(funcname)
        index = item.__dataindex__
        parent = item.__parent__
        type_hints = item.__type_hints__

        self.__configwidget = QtWidgets.QWidget()
        self.__layoutwidget = QtWidgets.QVBoxLayout(self.__configwidget)
        self.__layoutwidget.addWidget(QtWidgets.QLabel('Choose color for {:s}'.format(str(index))))
        self.__configwidget_input = QtWidgets.QColorDialog()
        self.__layoutwidget.addWidget(self.__configwidget_input)
        # Buttons
        self.__configwidget_apply = QtWidgets.QPushButton('Apply')
        self.__configwidget_apply.clicked.connect(self.applyGuiInput)
        self.__configwidget_apply.__configType = 'configColor'
        self.__configwidget_apply.item = item
        self.__layoutwidget.addWidget(self.__configwidget_apply)
        logger.debug('Done')

    def createConfigWidgetLiteral(self, item):
        logger.debug('createConfigWidgetLiteral')
        index = item.__dataindex__
        parent = item.__parent__
        type_hints = item.__type_hints__
        #
        if type_hints.__name__ == 'Literal':
            literal_options = typing.get_args(type_hints)
        else:
            literal_options = item.__literal_options__
        self.__configwidget = QtWidgets.QWidget()
        self.__layoutwidget = QtWidgets.QVBoxLayout(self.__configwidget)
        self.__layoutwidget.addWidget(QtWidgets.QLabel('Choose option for {:s}'.format(str(index))))
        self.__configwidget_input = QtWidgets.QComboBox()
        for o in literal_options:
            self.__configwidget_input.addItem(str(o))

        self.__layoutwidget.addWidget(self.__configwidget_input)
        # Buttons
        self.__configwidget_apply = QtWidgets.QPushButton('Apply')
        self.__configwidget_apply.clicked.connect(self.applyGuiInput)
        self.__configwidget_apply.__configType = 'configLiteral'
        self.__configwidget_apply.item = item
        self.__layoutwidget.addWidget(self.__configwidget_apply)



    def createConfigWidgetBaseModel(self, item):
        logger.debug('createConfigWidgetBaseModel')
        self.__configwidget = QtWidgets.QWidget()

    def createConfigWidgetList(self, item):
        logger.debug('createConfigWidgetList')
        self.createConfigWidgetDict(item)
        self.__layoutwidget.removeWidget(self.__keylabel)
        self.__layoutwidget.removeWidget(self.__configwidget_input)
        self.__keylabel.close()
        self.__configwidget_input.close()
        self.__configwidget_apply.__configType = 'configList'

    def createConfigWidgetBaseModelNew(self, item):
        logger.debug('createConfigWidgetBaseModelNew')
        data = item.__data__
        try:
            model_config = data.model_config['extra']
        except:
            model_config = 'omitted'

        if 'allow' in model_config.lower():
            editable = True
        else:
            editable = False

        if editable:
            self.createConfigWidgetDict(item)
            self.__configwidget_apply.__configType = 'configBaseModel'
            self.__configwidget_input.setText('newattribute')
        else:
            self.__configwidget = QtWidgets.QWidget()
            self.__layoutwidget = QtWidgets.QVBoxLayout(self.__configwidget)
            self.__layoutwidget.addWidget(QtWidgets.QLabel('Cannot add attribute'))

    def createConfigWidgetDict(self, item):
        funcname = __name__ + '.createConfigWidgetDict():'
        logger.debug('createConfigWidgetDict')
        index = item.__dataindex__
        parent = item.__parent__
        data = item.__data__
        type_hints_standard = typing.Union[bool, int, float, str]
        try:
            type_hints = item.__type_hints__
            if type_hints is None:
                logger.debug(funcname + 'Create standard type hints')
                type_hints = type_hints_standard
                item.__type_hints__ = type_hints_standard
        except:
            type_hints = type_hints_standard

        print('Type hints', type_hints)
        type_dict = self.interprete_type_hints(type_hints)
        print('Type dict', type_dict)
        addCombo = QtWidgets.QComboBox()
        has_combo = True
        logger.debug(funcname + 'Filling combo box')
        for iarg, argstr in enumerate(type_dict['type_args']):
            print('Argstr ...', argstr)
            # If annotated data is available look search for known datatypes
            if argstr == 'Annotated':
                # print('Annotated argument')
                annotations = typing.get_args(typing.get_args(item.__type_hints__)[iarg])
                # print('Annotations ...', annotations)
                for annotation in annotations:
                    if annotation == 'RedvyprAddressStr':
                        argstr = annotation
                        break
            addCombo.addItem(argstr)
            # Save also the object
            user_role_config = 11
            role = QtCore.Qt.UserRole + user_role_config
            dobject = type_dict['datatype_objects'][argstr]
            addCombo.setItemData(iarg, dobject, role)

        self.__configwidget_combo = addCombo
        self.__configwidget = QtWidgets.QWidget()
        self.__layoutwidget = QtWidgets.QFormLayout(self.__configwidget)
        keylabel = QtWidgets.QLabel('Datakey')
        self.__keylabel = keylabel
        self.__configwidget_input = QtWidgets.QLineEdit('newkey')
        #self.__configwidget_input.setText(str(data))
        #self.__layoutwidget.addRow(QtWidgets.QLabel('Add a new entry for {:s}'.format(str(index))))
        self.__layoutwidget.addRow(QtWidgets.QLabel('Add a new entry'))
        self.__layoutwidget.addRow(keylabel, self.__configwidget_input)
        self.__layoutwidget.addRow(QtWidgets.QLabel('Datatype'), addCombo)
        # Buttons
        self.__configwidget_apply = QtWidgets.QPushButton('Add')
        self.__configwidget_apply.clicked.connect(self.applyGuiInput)
        self.__configwidget_apply.__configType = 'configDict'
        self.__configwidget_apply.item = item
        self.__layoutwidget.addRow(self.__configwidget_apply)

    def createConfigWidgetStr(self, item):
        index = item.__dataindex__
        parent = item.__parent__
        data = item.__data__
        parentparent = parent.__parent__
        self.__configwidget = QtWidgets.QWidget()
        self.__layoutwidget = QtWidgets.QFormLayout(self.__configwidget)
        self.__configwidget_input = QtWidgets.QLineEdit()
        self.__configwidget_input.setText(str(data))
        self.__layoutwidget.addRow(QtWidgets.QLabel('Enter string for {:s}'.format(str(index))))
        self.__layoutwidget.addRow(QtWidgets.QLabel('Value'), self.__configwidget_input)
        # Buttons
        self.__configwidget_apply = QtWidgets.QPushButton('Apply')
        self.__configwidget_apply.clicked.connect(self.applyGuiInput)
        self.__configwidget_apply.__configType = 'configStr'
        self.__configwidget_apply.item = item
        self.__configwidget_cancel = QtWidgets.QPushButton('Cancel')
        self.__layoutwidget.addRow(self.__configwidget_apply)
        self.__layoutwidget.addRow(self.__configwidget_cancel)

    def createConfigWidgetBytes(self, item):
        index = item.__dataindex__
        parent = item.__parent__
        data = item.__data__
        parentparent = parent.__parent__
        self.__configwidget = QtWidgets.QWidget()
        self.__layoutwidget = QtWidgets.QFormLayout(self.__configwidget)
        self.__configwidget_input = QtWidgets.QLineEdit()
        self.__configwidget_input.setText(str(data))
        self.__layoutwidget.addRow(QtWidgets.QLabel('Enter string for {:s}'.format(str(index))))
        self.__layoutwidget.addRow(QtWidgets.QLabel('Value'), self.__configwidget_input)
        # Buttons
        self.__configwidget_apply = QtWidgets.QPushButton('Apply')
        self.__configwidget_apply.clicked.connect(self.applyGuiInput)
        self.__configwidget_apply.__configType = 'configBytes'
        self.__configwidget_apply.item = item
        self.__configwidget_cancel = QtWidgets.QPushButton('Cancel')
        self.__layoutwidget.addRow(self.__configwidget_apply)
        self.__layoutwidget.addRow(self.__configwidget_cancel)

    def createConfigWidgetRedvyprAddressStr(self, item):
        funcname = __name__ + 'createConfigWidgetRedvyprAddressStr():'
        logger.debug(funcname)
        try:
            index = item.__dataindex__
            parent = item.__parent__
            data = item.__data__
            data = str(data)
            print('Data data',data,type(data))
            print('Data data', item.text(1))
            parentparent = parent.__parent__
            self.__configwidget = QtWidgets.QWidget()
            self.__layoutwidget = QtWidgets.QFormLayout(self.__configwidget)
            if self.redvypr is not None:
                self.__configwidget_input = RedvyprAddressWidget(data,redvypr=self.redvypr)
            else:
                self.__configwidget_input = RedvyprAddressWidgetSimple(data)
            self.__configwidget_input.__configType = 'configRedvyprAddress'
            self.__configwidget_input.item = item
            self.__configwidget_input.address_finished.connect(self.applyGuiInput)
            self.__layoutwidget.addWidget(self.__configwidget_input)
        except:
            logger.info('Error in RedvyprAddress',exc_info=True)


    def createConfigWidgetDateTime(self, item, dateformat='yyyy-MM-dd HH:MM:ss'):

        index = item.__dataindex__
        parent = item.__parent__
        data = item.__data__
        parentparent = parent.__parent__
        self.__configwidget = QtWidgets.QWidget()
        self.__layoutwidget = QtWidgets.QFormLayout(self.__configwidget)
        self.__configwidget_input = QtWidgets.QDateTimeEdit(data)
        self.__configwidget_input.setDisplayFormat(dateformat)
        self.__layoutwidget.addRow(QtWidgets.QLabel('Enter date/time for {:s}'.format(str(index))))
        self.__layoutwidget.addRow(QtWidgets.QLabel('Date/Time'), self.__configwidget_input)
        # Buttons
        self.__configwidget_apply = QtWidgets.QPushButton('Apply')
        self.__configwidget_apply.clicked.connect(self.applyGuiInput)
        self.__configwidget_apply.__configType = 'configDateTime'
        self.__configwidget_apply.item = item
        self.__configwidget_cancel = QtWidgets.QPushButton('Cancel')
        self.__layoutwidget.addRow(self.__configwidget_apply)
        self.__layoutwidget.addRow(self.__configwidget_cancel)
        # if (parentparent is not None):  # Remove button
        #    self.__add_remove_btn__(self.__layoutwidget_int, item=item, dtype='str')
        #    try:
        #        removable = parent.__data__.children_removable
        #    except Exception as e:
        #        removable = True

        #    self.__configwidget_remove.setEnabled(removable)

    def createConfigWidgetBool(self, item):

        index = item.__dataindex__
        parent = item.__parent__
        data = item.__data__
        parentparent = parent.__parent__
        self.__configwidget = QtWidgets.QWidget()
        self.__layoutwidget = QtWidgets.QFormLayout(self.__configwidget)
        self.__configwidget_input = QtWidgets.QComboBox()
        self.__configwidget_input.addItem('True')
        self.__configwidget_input.addItem('False')
        self.__layoutwidget.addRow(QtWidgets.QLabel('Enter bool for {:s}'.format(str(index))))
        self.__layoutwidget.addRow(QtWidgets.QLabel('Value'), self.__configwidget_input)
        # Buttons
        self.__configwidget_apply = QtWidgets.QPushButton('Apply')
        self.__configwidget_apply.clicked.connect(self.applyGuiInput)
        self.__configwidget_apply.__configType = 'configBool'
        self.__configwidget_apply.item = item
        self.__configwidget_cancel = QtWidgets.QPushButton('Cancel')
        self.__layoutwidget.addRow(self.__configwidget_apply)
        self.__layoutwidget.addRow(self.__configwidget_cancel)

    def createConfigWidgetNumber(self, item, dtype='int'):
        """
        Creates a widgets to modify an integer value

        Returns:

        """
        funcname = __name__ + '.createConfigWidgetNumber():'
        logger.debug(funcname)
        index = item.__dataindex__
        parent = item.__parent__
        data = item.__data__
        parentparent = parent.__parent__
        self.__configwidget = QtWidgets.QWidget()
        self.__layoutwidget = QtWidgets.QFormLayout(self.__configwidget)
        #self.__configwidget_input = QtWidgets.QLineEdit()
        if(dtype=='int'):
            self.__configwidget_input = QtWidgets.QSpinBox()
            self.__configwidget_input.setRange(int(-1e9),int(1e9))
            try:
                value = int(item.text(1))
            except:
                value = 0
            self.__configwidget_input.setValue(value)
        else:
            self.__configwidget_input = QtWidgets.QDoubleSpinBox()
            self.__configwidget_input.setRange(-1e9, 1e9)
            try:
                value = float(item.text(1))
            except:
                value = 0.0
            self.__configwidget_input.setValue(value)

        self.__layoutwidget.addRow(QtWidgets.QLabel('Enter value for {:s}'.format(str(index))))
        self.__layoutwidget.addRow(QtWidgets.QLabel('Value'), self.__configwidget_input)

        # Buttons
        self.__configwidget_apply = QtWidgets.QPushButton('Apply')
        self.__configwidget_apply.clicked.connect(self.applyGuiInput)
        self.__configwidget_apply.__configType = 'configNumber'
        self.__configwidget_apply.item = item
        self.__configwidget_cancel = QtWidgets.QPushButton('Cancel')
        self.__layoutwidget.addRow(self.__configwidget_apply)
        self.__layoutwidget.addRow(self.__configwidget_cancel)


    def applyGuiInput(self):
        funcname = __name__ + '.applyGuiInput():'
        logger.debug(funcname)
        item = self.sender().item
        print('Some info')
        print('item_data',item.__data__)
        print('item_parent',item.__dataparent__)
        print('item_dataindex',item.__dataindex__)
        item_data = item.__data__
        # The flag to add a new item
        flag_add = item.__flag_add__

        print('datatypestr',item.__datatypestr__)
        #print(item.__parent__)
        print('Some info done')
        data_set = False
        if self.sender().__configType == 'configNumber':
            #print(funcname + ' ' + self.sender().__configType)
            #print('Setting data')
            #print('Reloading data')
            #print('Item',item)
            self.__configwidget_input
            data = self.__configwidget_input.value()  # Works for int/float spinboxes
            data_set = True

        elif self.sender().__configType == 'configList':
            logger.debug(funcname + 'Processing configList')
            data = item.__data__
            datakey = self.__configwidget_input.text()  # ComboBox
            index = self.__configwidget_combo.currentIndex()
            print('Dict newkey', datakey, 'Index', index)
            user_role_config = 11
            role = QtCore.Qt.UserRole + user_role_config
            dobject = self.__configwidget_combo.itemData(index, role)
            print('Dobject', dobject)
            dobject_add = dobject()
            data.append(dobject_add)
            # Reload and redraw all data
            self.configWidget.reload_data()
            self.config_changed_flag.emit()
            return

        elif self.sender().__configType == 'configBaseModel':
            logger.debug(funcname + 'Processing configBaseModel')
            data = item.__data__
            datakey = self.__configwidget_input.text()  # ComboBox
            index = self.__configwidget_combo.currentIndex()
            print('Dict newkey', datakey,'Index',index)
            user_role_config = 11
            role = QtCore.Qt.UserRole + user_role_config
            dobject = self.__configwidget_combo.itemData(index, role)
            print('Dobject',dobject)
            dobject_add = dobject()
            #data[datakey] = dobject_add
            setattr(data, datakey, dobject_add)
            # Reload and redraw all data
            self.configWidget.reload_data()
            self.config_changed_flag.emit()
            return

        elif self.sender().__configType == 'configDict':
            logger.debug(funcname + 'Processing dictionary')
            data = item.__data__
            datakey = self.__configwidget_input.text()  # ComboBox
            index = self.__configwidget_combo.currentIndex()
            print('Dict newkey', datakey,'Index',index)
            user_role_config = 11
            role = QtCore.Qt.UserRole + user_role_config
            dobject = self.__configwidget_combo.itemData(index, role)
            print('Dobject',dobject)
            dobject_add = dobject()
            data[datakey] = dobject_add
            # Reload and redraw all data
            self.configWidget.reload_data()
            self.config_changed_flag.emit()
            return

        elif self.sender().__configType == 'configLiteral':
            logger.debug(funcname + 'Processing configLiteral')
            data = self.__configwidget_input.currentText()  # ComboBox
            data_set = True

        elif self.sender().__configType == 'configColor':
            logger.debug(funcname + 'Processing configColor')
            #print(funcname + ' Color')
            color = self.__configwidget_input.currentColor()  # ComboBox
            #color1 = color.getRgbF()
            color1 = color.getRgb()
            color_tmp = (color1[0],color1[1],color1[2])
            rint = int(color1[0] * 255)
            gint = int(color1[1] * 255)
            bint = int(color1[2] * 255)
            data = pydColor(color_tmp)
            #print('Got color',data)
            data_set = True

        elif self.sender().__configType == 'configStr':
            logger.debug(funcname + 'Processing configStr')
            data = self.__configwidget_input.text()  # Textbox
            data_set = True

        elif self.sender().__configType == 'configBytes':
            logger.debug(funcname + 'Processing configBytes')
            data = self.__configwidget_input.text()  # Textbox
            print('data',data)
            if data.startswith("b'") and data.endswith("'"):
                data = eval(data)
            else:
                data = data.encode('utf-8')
            data_set = True

        elif self.sender().__configType == 'configRedvyprAddressStr':
            logger.debug(funcname + 'Processing configRedvyprAddressStr')
            #data = self.__configwidget_input.text()  # Textbox
            data = RedvyprAddress(self.__configwidget_input.text())  # Textbox
            data_set = True

        elif self.sender().__configType == 'configDateTime':
            logger.debug(funcname + 'Processing configDateTime')
            #print('Datetime')
            datetmp = self.__configwidget_input.dateTime()
            data = datetmp.toPyDateTime()
            data_set = True

        elif self.sender().__configType == 'configBool':
            logger.debug(funcname + 'Processing configBool')
            data = self.__configwidget_input.currentText() == 'True' # Combobox
            data_set = True

        elif self.sender().__configType == 'configRedvyprAddress':
            logger.debug(funcname + 'Processing configRedvyprAddress')
            addr = self.sender().redvypr_address
            #data = str(addr)
            data = addr
            data_set = True
        else:
            logger.warning('Unknown config type {}'.format(self.sender().__configType))

        if data_set:
            if True: # or an existing model was changed
                #print('Type',type(item.__dataparent__))
                # Dictionaries?!
                if pydantic.BaseModel in item.__dataparent__.__class__.__mro__:
                    logger.debug('Adding data to attribute "{}" of pydantic basemodel '.format(item.__dataparent__))
                    print('Data',data,type(data))
                    setattr(item.__dataparent__,item.__dataindex__, data)
                elif isinstance(item.__dataparent__, list):
                    logger.debug('Changing data at index {}'.format(item.__dataindex__))
                    item.__dataparent__[item.__dataindex__] = data
                elif isinstance(item.__dataparent__, dict):
                    logger.debug('Changing data at index of dict {}'.format(item.__dataindex__))
                    item.__dataparent__[item.__dataindex__] =  data
                else:
                    logger.warning('Could not add data')
                # item.setText(1, str(data_num))

            # Reload and redraw all data
            self.configWidget.reload_data()
            self.config_changed_flag.emit()



#
#
#
# pydanticQTreeWidget
#
#
class pydanticQTreeWidget(QtWidgets.QTreeWidget):
    """ Qtreewidget that display a pydantic object
    """

    def __init__(self, data=None, dataname='data', show_datatype=True, show_editable_only=True, exclude=[]):
        funcname = __name__ + '.__init__():'
        super().__init__()
        logger.debug(funcname + str(data))
        self.setExpandsOnDoubleClick(False)
        # make only the first column editable
        #self.setEditTriggers(self.NoEditTriggers)
        #self.header().setVisible(False)
        self.exclude = exclude
        self.setHeaderLabels(['Variable','Value','Type'])
        self.data = data
        self.dataname = dataname
        # Display options
        self.show_editable_only = show_editable_only
        # Create the root item
        self.root = self.invisibleRootItem()
        #self.root.__data__ = data
        self.root.__dataindex__ = ''
        self.root.__datatypestr__ = ''
        self.root.__parent__ = None
        self.setColumnCount(3)
        self.create_qtree()


        # Show the datatpye column
        if show_datatype == False:
            self.header().hideSection(2)
        #self.itemExpanded.connect(self.resize_view)
        #self.itemCollapsed.connect(self.resize_view)

    def seq_iter(self, obj):
        """
        To treat dictsionaries and lists equally this functions returns either the keys of dictionararies or the indices of a list.
        This allows a
        index = seq_iter(data)
        for index in data:
            data[index]

        with index being a key or an int.

        Args:
            obj:

        Returns:
            list of indicies

        """
        obj_test = obj

        if isinstance(obj_test, dict):
            return obj
        elif isinstance(obj_test, list):
            return range(0, len(obj))
        else:
            return None

    def reload_data(self, data=None):
        if data is not None:
            self.data = data

        self.clear()
        self.create_qtree()
        self.expandAll()
        self.resizeColumnToContents(0)

    def create_item(self, index, data, parent, edit_flag=None):
        """
        Creates recursively qtreewidgetitems. If the item to be created is a sequence (dict or list), it calls itself as often as it finds a real value
        Args:
            index:
            data:
            parent:

        Returns:

        """
        funcname = __name__ + '.create_item():'
        logger.debug(funcname)

        #print('Parent',parent)
        #print('Hallo',type(data))
        #print('Hallo2',data.__class__.__base__)
        flag_basemodel = False
        if isinstance(data, dict):
            #print('dict')
            flag_iterate = True
            flag_add_entry = True
        elif isinstance(data, list):
            #print('list')
            flag_iterate = True
            flag_add_entry = True
        elif pydantic.BaseModel in data.__class__.__mro__:
            #print('basemodel')
            try:
                model_config = data.model_config['extra']
            except:
                model_config = 'omitted'
            flag_iterate = True
            flag_basemodel = True
            flag_add_entry = False
            if 'allow' in model_config.lower():
                flag_add_entry = True
        else:
            #print('item')
            flag_iterate = False
            flag_add_entry = False




        # Find an optional "editable" flag
        editable = True
        try:
            #print('test editable ...')
            #print('Editable',parent.__data__)
            attr = getattr(parent.__data__,index)
            mfields = parent.__data__.model_fields[index]
            #print('Mfields ...',mfields)
            editable = mfields.json_schema_extra['editable']
            #editable = parent.__data__[index].json_schema_extra['editable']
            logger.debug('{} has an editable flag with {}'.format(index,editable))
            #print('done test editable ...')
        except:
            editable = True
            #logger.debug('extra fields {}'.format(index),exc_info=True)

        # Try to find a level flag
        level = 0
        try:
            # print('test editable ...')
            # print('Editable',parent.__data__)
            attr = getattr(parent.__data__, index)
            mfields = parent.__data__.model_fields[index]
            # print('Mfields ...',mfields)
            level = mfields.json_schema_extra['level']
            # editable = parent.__data__[index].json_schema_extra['editable']
            logger.debug('{} has an level with {}'.format(index, level))
            # print('done test editable ...')
        except:
            level = 0
            # logger.debug('extra fields {}'.format(index),exc_info=True)

        #print('editable', editable)
        # Get the parentdata
        try:
            parentdata = parent.__data__
        except:
            parentdata = None

        try:  # If the parent allow to add entries, this item is removable
            removable = parent.__flag_add_entry__
            #print('removable', removable)
        except:
            removable = False


        if self.show_editable_only == False:
            flag_show_item = True
        elif editable and self.show_editable_only:
            flag_show_item = True
        else:
            flag_show_item = False

        if flag_show_item:
            if True:
                data_value = data  #
                # Check for the types
                type_hints_index = None
                typestr = data_value.__class__.__name__
                datatype = data_value.__class__
                try:
                    # Check if the parent is a pydantic Basemodel child
                    if pydantic.BaseModel in parentdata.__class__.__mro__:
                        # Check if the item is an etxra field that can
                        # be removed or a predefined field
                        base_attributes = parentdata.model_construct().model_dump().keys()
                        if index in base_attributes:
                            logger.debug('Attribute {} is a predefined attribute, not removable'.format(index))
                            removable = False

                        type_hints = typing.get_type_hints(parentdata, include_extras=True)
                        # save the type hints
                        type_hints_index = type_hints[index]
                        #print('type hints index', type_hints_index)
                        typestr = type_hints[index].__name__
                        datatype = type_hints[index]
                        # If annotated data is available look search for known datatypes
                        if typestr == 'Annotated':
                            #print('Annotated')
                            annotations = typing.get_args(type_hints[index])
                            for annotation in annotations:
                                if annotation == 'RedvyprAddressStr':
                                    typestr = annotation
                                    break

                except:
                    logger.debug('Could not get type hints for {}'.format(index),exc_info=True)

                indexstr = str(index)
                # Check if item should be excluded
                if indexstr in self.exclude:
                    return

            if (flag_iterate == False):  # Check if we have an item that is something with data (not a pydantic module, list or dict)
                item = QtWidgets.QTreeWidgetItem([indexstr, str(data_value),typestr])
                #item.setFlags(item.flags() | QtCore.Qt.ItemIsEditable)  # editable
                item.__data__ = data
                item.__dataparent__ = parentdata# parent.__data__ # can be used to reference the data (and change it)
                item.__dataindex__ = index
                item.__datatypestr__ = typestr
                item.__datatype__ = datatype
                item.__parent__ = parent
                item.__type_hints__ = type_hints_index
                item.__flag_add_entry__ = flag_add_entry
                item.__removable__ = removable
                item.__editable__ = editable
                item.__level__ = level
                # Add the item to the data
                #print('data',data)
                #print('data',type(data))
                #data.__item__ = item

                index_child = self.item_is_child(parent, item)
                if index_child == None:  # Check if the item is already existing, if no add it
                    parent.addChild(item)
                else: # Update the data (even if it hasnt changed
                    parent.child(index_child).setText(1,str(data_value))

            else: # Item that is iterable and can be a parent
                #print('loop')
                datatmp = data
                typestr = datatmp.__class__.__name__
                flag_modifiable = True
                indexstr = str(index)
                # Check if item should be excluded
                if indexstr in self.exclude:
                    return

                #print('gf',index)
                #print('gf type', type(index))
                #print('Hallo',str(index))
                # Create new item
                newparent = QtWidgets.QTreeWidgetItem([str(index), '',typestr])
                item = newparent
                newparent.__data__ = datatmp
                newparent.__dataindex__ = index
                newparent.__datatypestr__ = typestr
                newparent.__datatype__ = datatype
                newparent.__parent__ = parent
                newparent.__dataparent__ = parentdata  # parent.__data__ # can be used to reference the data (and change it)
                newparent.__modifiable__ = flag_modifiable
                newparent.__type_hints__ = type_hints_index
                newparent.__flag_add_entry__ = flag_add_entry
                newparent.__removable__ = removable
                newparent.__editable__ = editable
                #try:
                #    newparent.__dataparent__ = parent.__data__  # can be used to reference the data (and change it)
                #except:
                #    newparent.__dataparent__ = None

                index_child = self.item_is_child(parent, newparent)
                if index_child == None:  # Check if the item is already existing, if no add it
                    parent.addChild(newparent)
                else:
                    newparent = parent.child(index_child)


                for numi, newindex in enumerate(data):
                    if isinstance(data, dict):
                        #print('Dict',newindex)
                        newdata = datatmp[newindex]
                    elif isinstance(data, list):
                        #print('List', newindex)
                        newdata = datatmp[numi]
                        newindex = numi
                    elif pydantic.BaseModel in data.__class__.__mro__:
                        #print('numi',numi)
                        #print('newindex', newindex)
                        newdata = newindex[1]
                        newindex = newindex[0]
                        #print('basemodel newdata',newdata)
                        #print('newindex',newindex)
                        #print('newparent',newparent)
                    else:
                        logger.warning('Cannot iterate over type {}',type(data))

                    self.create_item(newindex, newdata, newparent, edit_flag= edit_flag)

    def item_is_child(self,parent,child):
        """
        Checks if the item is a child already

        Args:
            parent:
            child:

        Returns:

        """
        numchilds  = parent.childCount()
        for i in range(numchilds):
            testchild = parent.child(i)
            #flag1 = testchild.__data__        == child.__data__
            flag1 = True
            flag2 = testchild.__dataindex__   == child.__dataindex__
            #flag3 = testchild.__datatypestr__ == child.__datatypestr__
            flag3 = True
            flag4 = testchild.__parent__      == child.__parent__

            #print('fdsfd',i,testchild.__data__,child.__data__)
            #print('flags',flag1,flag2,flag3,flag4)
            if(flag1 and flag2 and flag3 and flag4):
                return i

        return None

    def create_qtree(self, editable=True):
        """Creates a new qtree from the configuration and replaces the data in
        the dictionary with a configdata obejct, that save the data
        but also the qtreeitem, making it possible to have a synced
        config dictionary from/with a qtreewidgetitem. TODO: Worth to
        replace with a qviewitem?

        """
        funcname = __name__ + '.create_qtree():'
        logger.debug(funcname)
        self.blockSignals(True)
        #print('data',self.data)
        if True:
            self.create_item(self.dataname, self.data, self.root)

        self.dataitem = self.root.child(0)
        self.resizeColumnToContents(0)
        self.blockSignals(False)
        self.expandAll()
        self.resizeColumnToContents(0)



    def resize_view(self):
        pass
        #self.resizeColumnToContents(0)


class dictQTreeWidget(QtWidgets.QTreeWidget):
    """ Qtreewidget that display a Dict data structure
    """

    def __init__(self, data={}, dataname='data', show_datatype = True):
        funcname = __name__ + '.__init__():'
        super().__init__()
        if(type(data) == dict): # Convert to configDict to allow to store extra attributes
            pass
        else:
            raise TypeError(funcname + ' Expecting a dict or a configDict as data')
        logger.debug(funcname + str(data))
        self.setExpandsOnDoubleClick(False)
        # make only the first column editable
        #self.setEditTriggers(self.NoEditTriggers)
        #self.header().setVisible(False)
        self.setHeaderLabels(['Variable','Value','Type'])
        self.data     = data
        self.dataname = dataname
        # Create the root item
        self.root = self.invisibleRootItem()
        #self.root.__data__ = data
        self.root.__dataindex__ = ''
        self.root.__datatypestr__ = ''
        self.root.__parent__ = None
        self.setColumnCount(3)
        self.create_qtree()

        # Show the datatpye column
        if show_datatype == False:
            self.header().hideSection(2)
        #self.itemExpanded.connect(self.resize_view)
        #self.itemCollapsed.connect(self.resize_view)

    def reload_data(self,data):
        funcname = __name__ + '.reload_data():'
        #print(funcname)
        #print('data',data)
        #print('------')
        self.data = data
        self.clear()
        self.create_qtree()
        self.expandAll()
        self.resizeColumnToContents(0)

    def seq_iter(self, obj):
        """
        To treat dictsionaries and lists equally this functions returns either the keys of dictionararies or the indices of a list.
        This allows a
        index = seq_iter(data)
        for index in data:
            data[index]

        with index being a key or an int.

        Args:
            obj:

        Returns:
            list of indicies

        """
        if isinstance(obj, dict):
            return obj
        elif isinstance(obj, list):
            return range(0, len(obj))
        else:
            return None

    def create_item(self, index, data, parent):
        """
        Creates recursively qtreewidgetitems. If the item to be created is a sequence (dict or list), it calls itself as often as it finds a real value
        Args:
            index:
            data:
            parent:

        Returns:

        """
        sequence = self.seq_iter(data)
        if(sequence == None): # Check if we have an item that is something with data (not a list or dict)
            data_value = data  #
            typestr = data_value.__class__.__name__

            item       = QtWidgets.QTreeWidgetItem([str(index), str(data_value),typestr])
            #item.setFlags(item.flags() | QtCore.Qt.ItemIsEditable)  # editable
            item.__data__ = data
            item.__dataparent__   = parent.__data__ # can be used to reference the data (and change it)
            item.__dataindex__    = index
            item.__datatypestr__  = typestr
            item.__parent__       = parent
            # Add the item to the data
            #print('data',data)
            #print('data',type(data))

            index_child = self.item_is_child(parent, item)
            if  index_child == None:  # Check if the item is already existing, if no add it
                parent.addChild(item)
            else: # Update the data (even if it hasnt changed
                parent.child(index_child).setText(1,str(data_value))

        else:
            #print('loop')
            datatmp = data
            typestr = datatmp.__class__.__name__
            if(index is not None):
                indexstr = index
            newparent = QtWidgets.QTreeWidgetItem([str(index), '',typestr])
            item = newparent
            newparent.__data__ = datatmp
            newparent.__dataindex__ = index
            newparent.__datatypestr__ = typestr
            newparent.__parent__ = parent
            try:
                newparent.__dataparent__ = parent.__data__  # can be used to reference the data (and change it)
            except:
                newparent.__dataparent__ = None

            index_child = self.item_is_child(parent, newparent)
            if index_child == None:  # Check if the item is already existing, if no add it
                parent.addChild(newparent)
            else:
                newparent = parent.child(index_child)

            for newindex in sequence:
                newdata = datatmp[newindex]
                self.create_item(newindex,newdata,newparent)

    def item_is_child(self,parent,child):
        """
        Checks if the item is a child already

        Args:
            parent:
            child:

        Returns:

        """
        numchilds  = parent.childCount()
        for i in range(numchilds):
            testchild = parent.child(i)
            #flag1 = testchild.__data__        == child.__data__
            flag1 = True
            flag2 = testchild.__dataindex__   == child.__dataindex__
            #flag3 = testchild.__datatypestr__ == child.__datatypestr__
            flag3 = True
            flag4 = testchild.__parent__      == child.__parent__

            #print('fdsfd',i,testchild.__data__,child.__data__)
            #print('flags',flag1,flag2,flag3,flag4)
            if(flag1 and flag2 and flag3 and flag4):
                return i

        return None

    def create_qtree(self, editable=True):
        """Creates a new qtree from the configuration and replaces the data in
        the dictionary with a configdata obejct, that save the data
        but also the qtreeitem, making it possible to have a synced
        config dictionary from/with a qtreewidgetitem.

        """
        funcname = __name__ + '.create_qtree():'
        logger.debug(funcname)
        self.blockSignals(True)
        if True:
            self.create_item(self.dataname,self.data,self.root)

        self.dataitem = self.root.child(0)
        self.resizeColumnToContents(0)
        self.blockSignals(False)



    def resize_view(self):
        pass
        #self.resizeColumnToContents(0)



class datastreamMetadataWidget(datastreamQTreeWidget):
    def __init__(self, *args, metadata=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.metadata = metadata
        self.devicelist.itemClicked.connect(self.__item_clicked)

        # Create a splitter
        #if self.config_location == 'bottom':
        #    sdir = QtCore.Qt.Vertical
        #else:
        #    sdir = QtCore.Qt.Horizontal
        self.metadata_configWidget = QtWidgets.QWidget()
        self.metadata_configWidget_layout =  QtWidgets.QVBoxLayout(self.metadata_configWidget)
        self.radioMerged = QtWidgets.QRadioButton("Merged")
        self.radioMerged.toggled.connect(self.__metadata_mode_changed)
        self.radioExpanded = QtWidgets.QRadioButton("Expanded")
        self.radioLayout = QtWidgets.QHBoxLayout()
        self.radioLayout.addWidget(self.radioMerged)
        self.radioLayout.addWidget(self.radioExpanded)
        self.metadata_configWidget_layout.addLayout(self.radioLayout)
        self.radioMerged.setChecked(True)

        sdir = QtCore.Qt.Horizontal
        self.splitterMetadata = QtWidgets.QSplitter(sdir)
        self.layout.removeWidget(self.splitter)
        self.splitterMetadata.addWidget(self.splitter)
        self.splitterMetadata.addWidget(self.metadata_configWidget)
        #self.layout.removeWidget(self.deviceWidget)
        self.layout.addWidget(self.splitterMetadata,0,0)
        # Lets click an item
        item_select = self.devicelist.topLevelItem(self.devicelist.topLevelItemCount() - 1)
        if item_select is None:
            pass
        else:
            item_select.setSelected(True)
            self.__item_clicked(item_select, 0)



    def __metadata_mode_changed(self):
        funcname = __name__ + '.__metadata_mode_changed():'
        logger.debug(funcname)
        try:
            item = self.__item_selected
        except:
            item = self.devicelist.topLevelItem(self.devicelist.topLevelItemCount() - 1)

        self.__item_clicked(item, 0)
    def __item_clicked(self,item,col):
        """
        Called when an item in the qtree is clicked
        """
        funcname = __name__ + '__item_clicked():'
        logger.debug(funcname)
        self.__item_selected = item
        if item is None:
            logger.warning('Item is None, doing nothing')
        else:

            if self.radioMerged.isChecked():
                metadata_mode = 'merge'
            else:
                metadata_mode = 'dict'
            funcname = __name__ + '__item_clicked()'
            logger.debug(funcname)
            print('Item',item)
            try:
                print('Address1',item.raddress)
            except:
                logger.info('Could not get address',exc_info=True)
                pass

            raddress = item.raddress
            address_format = '/h/d/i/k'
            # This needs to be considered, what we actually want to get ...
            fstr1 = raddress.get_expand_explicit_str(address_format=address_format)
            raddress_metadata = RedvyprAddress(fstr1)
            print('Raddress_metadata',raddress_metadata)
            metadata = self.device.get_metadata(raddress_metadata,mode=metadata_mode)
            print('Got metadata',metadata)

            metadata_work = copy.deepcopy(metadata)
            if metadata_mode == 'dict':
                try:
                    metadata_work[raddress.get_str(address_format)]
                except:
                    metadata_work[raddress.get_str(address_format)] = {}
            else:
                metadata_work = {raddress.get_str(address_format):metadata_work}

            metadata_pydantic = RedvyprMetadataGeneral(address=metadata_work)
            #print('Metadata', metadata_pydantic)
            #print('Got metadata (keys)', metadata.keys())
            self.create_metadata_widget(metadata_pydantic,raddress)

    def create_metadata_widget(self,metadata, raddress):
        funcname = __name__ + '.create_metadata_widget():'
        logger.debug(funcname)
        #metadata_pydantic = RedvyprDeviceMetadata(**metadata)
        print('Metadata edit widget', metadata)
        self.__metadata_edit = metadata
        self.__metadata_address = raddress
        try:
            self.metadata_configWidget_layout.removeWidget(self.metadata_config)
        except:
            pass
        self.metadata_config = pydanticConfigWidget(metadata, configname='Metadata', close_after_editing=False)
        self.metadata_configWidget_layout.addWidget(self.metadata_config)
        self.metadata_config.config_editing_done.connect(self.metadata_config_apply)
        #self.metadata_config.show()


    def metadata_config_apply(self):
        funcname = __name__ + '.metadata_config_apply():'
        try:
            print('Apply')
            metadata = self.__metadata_edit
            print('Hallo',metadata)
            for address in metadata.address.keys():
                print('Updating metadata data',metadata.address[address])
                print('Updating metadata', address)
                if len(metadata.address[address].keys()) > 0:
                    self.device.set_metadata(address, metadata.address[address])
                else:
                    logger.debug('Not enough datakey so send')

        except:
            logger.info(funcname + 'Could not update metadata',exc_info=True)





import copy
import datetime
import logging
import queue
from PyQt5 import QtWidgets, QtCore, QtGui
import time
import numpy as np
import serial
import serial.tools.list_ports
import logging
import sys
import threading
import pydantic
import typing
from redvypr.data_packets import check_for_command, create_datadict
from redvypr.device import RedvyprDevice
import redvypr.files as redvypr_files
import redvypr.devices.plot.plot_widgets as redvypr_plot_widgets

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('manual')
logger.setLevel(logging.DEBUG)

redvypr_devicemodule = True

class DeviceBaseConfig(pydantic.BaseModel):
    publishes: bool = True
    subscribes: bool = True
    description: str = 'Allows to send data manually'
    gui_icon: str = 'mdi.serial-port'


def start(*args,**kwargs):
    return

class RedvyprDeviceWidget(QtWidgets.QWidget):
    def __init__(self,device=None):
        super(QtWidgets.QWidget, self).__init__()
        layout = QtWidgets.QGridLayout(self)
        self.device   = device


        self.sendbutton = QtWidgets.QPushButton("Send")
        self.sendbutton.clicked.connect(self.send_clicked)
        self.datakey_edit = QtWidgets.QLineEdit()
        self.datakey_edit.setText('comment')
        self.comment_edit = QtWidgets.QLineEdit()
        self.comment_edit.setText('Some manual comment')

        self.date_edit = QtWidgets.QDateTimeEdit()
        try:
            self.date_edit.setDateTime(datetime.datetime.now())
            timeformat = 'dd.MM.yyyy HH:MM:ss'
            self.date_edit.setDisplayFormat(timeformat)
        except:
            logger.info('Could net setup date',exc_info=True)
        self.datakey_date_edit = QtWidgets.QLineEdit()
        self.datakey_date_edit.setText('t')
        # Some numbers
        self.datakey_number_edit = QtWidgets.QLineEdit()
        self.datakey_number_edit.setText('data')
        self.number_edit = QtWidgets.QDoubleSpinBox()

        layout.addWidget(QtWidgets.QLabel("Datakey"),0,0)
        layout.addWidget(QtWidgets.QLabel("Data"),0,2)
        layout.addWidget(self.datakey_edit,layout.rowCount()+1,0,1,2)
        layout.addWidget(self.comment_edit,layout.rowCount()-1,2,1,5)
        layout.addWidget(self.datakey_date_edit, layout.rowCount()+1, 0, 1, 2)
        layout.addWidget(self.date_edit, layout.rowCount()-1, 2, 1, 5)
        layout.addWidget(self.datakey_number_edit, layout.rowCount()+1, 0, 1, 2)
        layout.addWidget(self.number_edit, layout.rowCount()-1, 2, 1, 5)

        layout.addWidget(self.sendbutton,layout.rowCount()+1,0,1,-1)
        layout.setRowStretch(layout.rowCount(), 1)

    def send_clicked(self):
        funcname = __name__ + '.send_clicked()'
        logger.debug(funcname)
        data = {}
        # Add the comment
        datakey = str(self.datakey_edit.text())
        data[datakey] = str(self.comment_edit.text())
        # Add the date
        datakey = str(self.datakey_date_edit.text())
        tsend = self.date_edit.dateTime().toPyDateTime()
        data[datakey] = tsend.timestamp()

        # Add the number
        datakey = str(self.datakey_number_edit.text())
        data[datakey] = self.number_edit.value()

        logger.debug('Sending data: {}'.format(data))
        self.device.dataqueue.put(data)



"""
Tests redvypr address creation and comparison with a data packet
"""
import time
import redvypr.data_packets
from redvypr.redvypr_address import RedvyprAddress

# Create some data
datapacket_raw = {'t1':1,'t2':10,'somestring':'Hello World','somelist':[1,2,5,7],'somedict':{'test':'ok','number':3.0}}
datapacket = datapacket_raw.copy()
# Create a bogus redvypr host
print('Creating hostinfo')
hostinfo = redvypr.create_hostinfo(hostname='someredvypr')
devicename = 'somedevice'
devicemodulename = 'somedevicemodulename'
numpacket = 0
tread = time.time() # The time the packet was received from the queue (in redvypr.distribute_data)
redvypr.packet_statistic.treat_datadict(datapacket, devicename, hostinfo, numpacket, tread,devicemodulename)
print('Datapacket', datapacket)

print('Redvypr address test')
rc0 = RedvyprAddress(datakey='t1')
# Comparisons
rc1 = RedvyprAddress(compare='["t1"]>=3')
rc2 = RedvyprAddress(compare='["t2"]>=3')
rc3 = RedvyprAddress(compare='["somelist"][2]<=1000')
print(rc1)

print('Testing comparisons')
print('Testing rc0',rc0)
test0 = datapacket in rc0
print('Test0',test0)
print('-----------------')
print('Testing rc1',rc1)
test1 = datapacket in rc1
print('Test1',test1)
print('-----------------')
print('Testing rc2',rc2)
test2 = datapacket in rc2
print('Test2',test2)
print('-----------------')
print('Testing rc3',rc3)
test3 = datapacket in rc3
print('Test3',test3)
print('-----------------')

print('Testing regular expressions')
# Regular expressions
re1 = RedvyprAddress('/d:{.*}')
re2 = RedvyprAddress('/d:{some.*}')
re3 = RedvyprAddress('/d:{other.*}')
reall = [re1,re2,re3]

for retest in reall:
    print('-----------------')
    print('Testing',retest)
    test = datapacket in retest
    print('Datapacket in test',test)
    print('-----------------')



print('Testing datakey comparisons of eval strings that are formatted differently')
rev0 = RedvyprAddress('/k:a')
rev1 = RedvyprAddress('/k:["a"][0]["somedata"]')
reeval_test = []
reeval_test.append(RedvyprAddress('/k:["a"]'))
reeval_test.append(RedvyprAddress('''/k:['a']'''))
reeval_test.append(RedvyprAddress("/k:['''a''']"))
reeval_test.append(RedvyprAddress('/k:{[a]}'))
reeval_test.append(RedvyprAddress("/k:['''a'''][0]['somedata']"))
reeval_test.append(RedvyprAddress("/k:['''a'''][1]['somedata']"))

print('----- test 1 -----')
for rev_test in reeval_test:
    print('{} in {}: {}'.format(rev_test,rev0,rev_test in rev0))
print('----- test 2 -----')
for rev_test in reeval_test:
    print('{} in {}: {}'.format(rev_test,rev1,rev_test in rev1))




from . import plot
#from . import plot_legacy
#from . import plot_widgets
from . import XYplotWidget
from . import XYPlotDevice
from . import rawdatadisp
from . import PcolorPlotDevice
redvypr_devicemodule = True


"""

The most simple test device

"""


import datetime
import logging
import queue
from PyQt5 import QtWidgets, QtCore, QtGui
import time
import numpy as np
import logging
import sys
import threading
import copy
from redvypr.device import RedvyprDeviceCustomConfig, RedvyprDevice
import redvypr.data_packets
from redvypr.data_packets import check_for_command
import pydantic


logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('calibration_autocal_test')
logger.setLevel(logging.DEBUG)

redvypr_devicemodule = True
description: str = 'A test device for the autocalibration feature of the the calibration device'
def start(device_info, config=None, dataqueue=None, datainqueue=None, statusqueue=None):
    funcname = __name__ + '.start():'
    logger.debug(funcname)
    print('Config',config)
    temp_set = 20.0
    temp = 10.0
    t_lowpass = 5
    dt_wait = 0.5
    i = 0
    counter = 0
    while True:
        try:
            data = datainqueue.get(block = False)
        except:
            data = None
        if(data is not None):
            command = check_for_command(data, thread_uuid=device_info['thread_uuid'])
            logger.debug('Got a command: {:s}'.format(str(data)))
            if (command == 'stop'):
                logger.debug('Command is for me: {:s}'.format(str(command)))
                break
            elif (command == 'set'):
                logger.debug('Command is for me: {:s}'.format(str(command)))
                temp_set = data['temp']

        dT = temp - temp_set
        temp = temp - dT * dt_wait/t_lowpass
        print('Temp',temp,'dT',dT,dT * dt_wait/t_lowpass)
        data = {}#redvypr.data_packets.create_datadict(device = device_info['device'])
        data['temp'] = temp
        data['temp_set'] = temp_set
        if abs(temp - temp_set)<0.001:
            data['temp_steady'] = 1
        else:
            data['temp_steady'] = 0

        print('Data',data)
        dataqueue.put(data)
        time.sleep(dt_wait)







"""
register sensors
 - check if data fits (in start), check_valid_data
 - if yes process_data, gives redvyprData
 - display data (widget)
"""

import datetime
import numpy as np
import logging
import sys
import threading
import copy
import yaml
import json
import typing
import pydantic
from PyQt5 import QtWidgets, QtCore, QtGui
from redvypr.device import RedvyprDevice
from redvypr.data_packets import check_for_command
#from redvypr.packet_statistics import get_keys_from_data
#import redvypr.packet_statistic as redvypr_packet_statistic
import redvypr.data_packets as data_packets
import redvypr.gui as gui
#import redvypr.config as redvypr_config
from redvypr.redvypr_address import RedvyprAddress
from redvypr.devices.plot import plot_widgets
from redvypr.devices.plot import XYplotWidget
import redvypr.files as redvypr_files
from redvypr.devices.sensors.calibration.calibration_models import calibration_HF, calibration_NTC, get_date_from_calibration
from .heatflow_sensors import parse_HFV_raw, process_IMU_packet, process_HFS_data, process_HF_data, DHFSWidget, HFVWidget, HFVWidget_config, sensor_DHFS50, logger_HFV4CH
from .temperature_array_sensors import process_TAR_data, sensor_TAR, TARWidget, TARWidget_config
from .sensorWidgets import sensorCoeffWidget, sensorConfigWidget

_icon_file = redvypr_files.icon_file

redvypr_devicemodule = True

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger('csvsensors')
logger.setLevel(logging.DEBUG)


class DeviceBaseConfig(pydantic.BaseModel):
    publishes: bool = True
    subscribes: bool = True
    description: str = 'Processing of temperature and heatflow sensordata'




class DeviceCustomConfig(pydantic.BaseModel):
    sensorconfigurations: typing.Dict[str,typing.Annotated[typing.Union[sensor_DHFS50,logger_HFV4CH, sensor_TAR], pydantic.Field(discriminator='sensor_type')]] = pydantic.Field(default={},description='Configuration of sensors, keys are their serial numbers')
    calibrations: typing.List[typing.Annotated[typing.Union[calibration_NTC, calibration_HF], pydantic.Field(discriminator='calibration_type')]] = pydantic.Field(default=[], description = 'List of sensor calibrations')
    calibration_files: list = pydantic.Field(default=[])
    datakeys: list =  pydantic.Field(default = ['data'], description = 'The datakeys to be looked for csv data')






packettypes = ['HF','HFS','HFV','IMU','TAR'] # This should be specified within the logger templates, work for later


def start(device_info, config = None, dataqueue = None, datainqueue = None, statusqueue = None):
    funcname = __name__ + '.start():'
    logger.debug(funcname)
    config = DeviceCustomConfig.model_validate(config)
    macs_found_HF = []  # List of all mac addresses found
    macs_found_HFSI = []  # List of all mac addresses found
    macs_found_IMU = []  # List of all mac addresses found
    macs_found = {'IMU':[],'HFSI':[],'HF':[],'TAR':[]}  # List of all mac addresses found
    avg_objs = {}
    print('Device info',device_info)
    #try:
    #    calfile = open(config['calibration_file'],'r')
    #    calibrations = yaml.safe_load(calfile)
    #except Exception as e:
    #    calibrations = None
    #    logger.exception(e)

    try:
        sensorconfigurations = config.sensorconfigurations
    except Exception as e:
        sensorconfigurations = None
        logger.exception(e)

    print('Logger configurations',sensorconfigurations)
    i = 0
    while True:
        data = datainqueue.get(block = True)
        if data is not None:
            #print('Got data',data)
            command = check_for_command(data,thread_uuid=device_info['thread_uuid'])
            #print('Got command',command, data)
            if (command == 'stop'):
                logger.debug('Command is for me: {:s}'.format(str(command)))
                statusqueue.put('stop')
                break
            elif (command == 'sensorconfigurations'):
                #config['sensorconfigurations'] = data['sensorconfigurations']
                logger.debug(funcname + ' sensorconfigurations updated (sensorconfigurations)')
            elif (command == 'config'):
                config = DeviceCustomConfig.model_validate(data['config'])
                logger.debug(funcname + ' sensorconfigurations updated (config)')


            #print('Got data',i,data)
            for datakey in config.datakeys:
                try:
                    hfdata = data[datakey]
                    break
                except:
                    hfdata = None
                    pass

            if hfdata is not None:
                if type(hfdata) == bytes:
                    try:
                        hfdatastr = hfdata.decode('utf-8')
                    except:
                        hfdatastr = 'X'
                elif type(hfdata) == str:
                    hfdatastr = hfdata
                else:
                    hfdatastr = 'X'
                        

                # Loop over all lines and check if we have a $ as start
                for dataline in hfdatastr.split('\n'):
                    valid_package = False
                    if len(dataline) > 0:
                        if dataline[0] == "$":
                            datasplit = dataline.split(',')
                            if len(datasplit) > 2:
                                sn = datasplit[0][1:]
                                # Check for a valid mac
                                try:
                                    macstr = datasplit[0][1:17]
                                    macint = int(macstr,16)
                                    logger.debug(funcname + 'Valid MAC: {:s}'.format(str(hex(macint))))
                                except:
                                    macint = None
                                    macstr = None
                                    logger.debug(funcname + 'Not valid MAC', exc_info = True)

                                packettype = datasplit[1]
                                #datapacket = None
                                #datapacket_HFSI = None
                                # Try to parse datapackets
                                print('Packettype',packettype)
                                if macstr is not None:
                                    # Check if the logger is already existing
                                    try:
                                        sensorconfig = config.sensorconfigurations[sn]
                                    except:
                                        logger.debug(funcname + ' New sensor', exc_info=True)
                                        statusmessage = {'status': 'newlogger', 'sn': sn, 'packettype': packettype,'dataline':dataline,'data':data}
                                        statusqueue.put(statusmessage)
                                        #print('new logger statusmessage', statusmessage)
                                        sensorconfig = None

                                    if packettype == 'IMU':
                                        try:
                                            datapacket = process_IMU_packet(dataline, data, macstr, macs_found)
                                            dataqueue.put(datapacket)
                                        except:
                                            logger.debug(' Could not decode {:s}'.format(str(dataline)), exc_info=True)
                                            datapacket = None

                                    elif packettype == 'TAR':
                                        #print('Parsing TAR')
                                        if sensorconfig is not None:
                                            try:
                                                datapackets_TAR = process_TAR_data(dataline, data, device_info, sensorconfig)
                                            except:
                                                logger.info(' Could not process data {:s}'.format(str(dataline)), exc_info=True)
                                                datapackets_TAR = []

                                            for datapacket_TAR in datapackets_TAR:
                                                dataqueue.put(datapacket_TAR)

                                    elif packettype == 'HFS':
                                        # Heatflow data in physical units
                                        try:
                                            datapacket_HFS = process_HFS_data(dataline, data, device_info)
                                            dataqueue.put(datapacket_HFS)
                                        except:
                                            logger.debug(' Could not decode {:s}'.format(str(dataline)), exc_info=True)
                                            datapacket_HFS = None

                                    elif packettype == 'HF':
                                        # Heatflow data in raw units
                                        print('HF packet')
                                        try:
                                            datapackets_HF = process_HF_data(dataline, data, device_info, sensorconfig, config, macs_found)
                                            for datapacket_HF in datapackets_HF:
                                                dataqueue.put(datapacket_HF)
                                        except:
                                            logger.debug(' Could not decode {:s}'.format(str(dataline)), exc_info=True)
                                            datapacket_HF = None

                                    elif packettype == 'HFV': # Voltage packet from a raspberry pi board
                                        #print('HFV')
                                        try:
                                            datapacket = parse_HFV_raw(dataline)
                                            #datapacket['t'] = data['t']
                                            datapacket['devicename'] = 'HFV_raw_' + sn
                                            # Check if there is a conversion to be done
                                            #datapacket_HFVSI = convert_HF(datapacket, calibrations)
                                            #datapacket_HFVSI['devicename'] = 'HFV_SI_' + sn
                                            #print('HALLO',config['sensorconfigurations'][sn]['channels'].keys())
                                            # Check if the logger is already existing
                                            try:
                                                config.sensorconfigurations[sn]
                                                flag_convert = True
                                            except:
                                                logger.debug(funcname + ' New logger.',exc_info=True)
                                                statusqueue.put({'status':'newlogger','sn':sn,'packettype':'HFV'})
                                                #print('datapacket',datapacket)
                                                flag_convert = False

                                            if flag_convert:
                                                flag_have_coeff = False
                                                #print('Converting data')
                                                try:
                                                    datapacket_HFSI = {'devicename':'HFV_SI_' + sn}
                                                    datapacket_HFSI['sn'] = datapacket['sn']
                                                    datapacket_HFSI['t'] = datapacket['t']
                                                    for channelconfig in config.sensorconfigurations[sn].channels:
                                                        channel = channelconfig[0]
                                                        calibration = channelconfig[1]
                                                        sn_sensor = calibration.sn
                                                        coeff = calibration.coeff
                                                        #print('Channelconfig', channelconfig, sn_sensor, coeff)
                                                        if len(sn_sensor) > 0:
                                                            #print('sensor',sn_sensor)
                                                            # This needs to be done more general ... checking units etc.
                                                            data_conv = datapacket[channel] * coeff * 1000
                                                            datapacket_HFSI[channel] = data_conv
                                                            datapacket_HFSI['type'] = 'HFVSI'
                                                            flag_have_coeff = True
                                                            # create a sn out of the logger and the attached sensor
                                                            sn_combined = datapacket_HFSI['sn'] + '_' + sn_sensor
                                                            # Check if the logger/sensor combinationis already known, otherwise send keyinfo
                                                            if sn_combined in macs_found['HFSI']:
                                                                pass
                                                            else:
                                                                logger.debug(
                                                                    'New HFV logger {:s}'.format(datapacket_HFSI['sn']))
                                                                macs_found_HFSI.append(datapacket_HFSI['sn'])
                                                                data_packets.add_metadata2datapacket(datapacket_HFSI,
                                                                                                     datakey=channel,
                                                                                                     unit='W m-2',
                                                                                                     description=None,
                                                                                                     infokey='sn',
                                                                                                     info=sn_sensor)
                                                                data_packets.add_metadata2datapacket(datapacket_HFSI,
                                                                                                     datakey=channel,
                                                                                                     infokey='sensortype',
                                                                                                     info=calibration.sensor_model)

                                                                macs_found['HFSI'].append(sn_combined)


                                                    if flag_have_coeff == False:
                                                        datapacket_HFSI = None


                                                except Exception as e:
                                                    coeff = None
                                                    logger.debug('Conversion error',exc_info=True)

                                        except Exception as e:
                                            logger.debug(' Could not decode {:s}'.format(str(dataline)))
                                            datapacket = None

                                        # print('HF packet',datapacket)
                                else:
                                    logger.debug(funcname + 'unknown packettype {:s}'.format(packettype))

class Device(RedvyprDevice):
    """
    heatflow_serial device
    """
    newlogger_signal = QtCore.pyqtSignal(dict)  # Signal with the new logger
    newcalibration_signal = QtCore.pyqtSignal()  # Signal for a new calibration
    newloggerconfig_signal = QtCore.pyqtSignal(dict)  # Signal when a new loggerconfiguration was applied
    def __init__(self, **kwargs):
        """
        """
        funcname = __name__ + '__init__()'
        super(Device, self).__init__(**kwargs)
        logger.debug(funcname)

        # Fill the calibration models
        calhints = typing.get_type_hints(self.custom_config)['calibrations']
        self.calibration_models = typing.get_args(typing.get_args(calhints)[0])

        #configtest = redvypr_config.dict_to_configDict(sensorconfig)
        #self.config['sensorconfigurations'][configtest['sn']] = configtest
        #for i,s in enumerate(self.config['sensors'].data):
        #    print('Subscribing to sensor',s)
        #    self.config['datastreams'].data.Device(append(redvypr.config.configString(''))
        #    self.subscribe_address(s)
        self.nosensorname = '<no sensor>'
        try:
            self.custom_config.calibrations[0]
        except:
            logger.debug(funcname + ' Will add <no sensor>')
            nosensor = calibration_HF()
            nosensor.sn = self.nosensorname
            #print('nosensorconfig',nosensorconfig)
            self.custom_config.calibrations.append(nosensor)
        # Sensordata
        self.sensordata_raw = {} # The datapackets
        self.sensordata = {} # The data by mac/parameter
        self.calfiles_processed = []

        # Add buffer for sensorconfigurations
        for sn in self.custom_config.sensorconfigurations.keys():
            self.sensordata[sn] = {'__np__': 0, '__raddr__': None, '__devices__': []}
            self.sensordata_raw[sn] = []
        # Check if there are already calibrations saved, and if yes, add the original files as read
        # Here a md5sum check could be done
        #for sn in self.config['calibrations'].keys():
        #    for cal in self.config['calibrations'][sn]:
        #        fname = cal['original_file']
        #        self.calfiles_processed.append(fname)
        self.populate_calibration()

    def get_sensor_parameter(self, sensor):
        funcname = __name__ + '.get_sensor_parameter():'
        logger.debug(funcname)
        parameter = sensor.parameter
        def __get_parameter_recursive__(index, parameter, parameter_all):
            for i, ptmp in enumerate(parameter):
                if isinstance(ptmp, list):
                    index_new = index + '[{}]'.format(i)
                    __get_parameter_recursive__(index_new, ptmp, parameter_all)
                # A pydantic parameter object
                elif ptmp.__class__.__base__ == pydantic.BaseModel:
                    if isinstance(parameter, list):
                        name = '{}[{}]'.format(index,i)
                    else:
                        name = str(i)
                    parameter_all.append({'parameter': ptmp, 'name': name,'parent':parameter,'index':i})


        parameter_all = []
        # Parameter should be a pydantic class
        if parameter.__class__.__base__ == pydantic.BaseModel:
            # Loop over all attributes
            for ptmp in parameter:
                par1 = ptmp[0]
                ptmp_child = getattr(parameter, par1)
                if isinstance(ptmp_child, list):
                    __get_parameter_recursive__(par1, ptmp_child, parameter_all)
                # A pydantic parameter object
                elif ptmp_child.__class__.__base__ == pydantic.BaseModel:
                    parameter_all.append({'parameter':ptmp_child,'name':par1,'parent':parameter,'index':par1})


                #print('ptmp',ptmp)

        else:
            raise Exception('parameter should be an instance of pydantic.BaseModel')


        return parameter_all

    def find_calibrations_for_sensor(self, sensor, match={'parameter':None, 'sn':None}):
        funcname = __name__ + '.find_calibrations_for_sensor():'
        logger.debug(funcname)
        parameter_all = self.get_sensor_parameter(sensor)
        for parameter_dict in parameter_all:
            logger.debug(funcname + 'Find calibration for {}'.format(parameter_dict['name']))
            cals = self.find_calibration_for_parameter(parameter_dict, sensor, match)
            if cals is not None:
                logger.debug(funcname + 'Calibration found')
                # cals is a list of calibrations sorted by date
                calibration = cals[0]
                #print('Setting calibration',calibration)
                self.set_calibration_for_parameter(parameter_dict, calibration)
            else:
                logger.debug(funcname + 'No calibration found')

    def find_calibration_for_parameter(self, parameter_dict, sensor, match={'parameter':None, 'sn':None}):
        """ Searches through calibrations to find a best match for the parameter.
        """
        funcname = __name__ + 'find_calibration_for_parameter():'
        logger.debug(funcname)
        cals = self.custom_config.calibrations
        #print('Parameter dict',parameter_dict)
        if True:
            para = parameter_dict['parameter']
            para_parent = parameter_dict['parent']
            para_name = parameter_dict['name']
            para_index = parameter_dict['index']
            logger.debug(funcname + 'Searching for calibration for parameter {}'.format(para_name))
            #print('Para', para_index, para)
            cal_match = []
            cal_match_date = []
            for cal in cals:
                logger.debug(funcname + 'Scanning calibration SN:{}, date: {}, ID: {}'.format(cal.sn, cal.date,cal.calibration_id))
                match_all = True
                for m in match.keys():
                    mcontent = match[m]
                    if m == 'sn' and mcontent is None: # Compare sn of sensor with sn of calibration
                        mcal = getattr(cal, m)
                        mpara = sensor.sn
                    else:
                        mcal = getattr(cal, m)
                        mpara = getattr(para, m)

                    # Check if the data in the dictionary shall be compared, of the content of the parameter
                    if mcontent is None:
                        mcompare = mpara
                    else:
                        mcompare = mcontent

                    print('match', m, 'mcal', mcal, 'mcompare', mcompare, mcal != mcompare)
                    if mcal != mcompare:
                        match_all = False

                if match_all:
                    logger.debug(funcname + 'Found matching parameter')
                    logger.debug(funcname + 'Calibration: {}'.format(cal))
                    cal_match.append(cal)
                    td = datetime.datetime.strptime(cal.date, '%Y-%m-%d %H:%M:%S.%f')
                    cal_match_date.append(td)
                    # Here the list is sorted by date

            # Return the calibrations if found, otherwise None
            if len(cal_match) > 0:
                # Sort the list according to the date
                cal_match_sorted = sorted(cal_match, reverse=True, key=lambda x: datetime.datetime.strptime(x.date, '%Y-%m-%d %H:%M:%S.%f'))
                return cal_match_sorted
            else:
                return None

    def set_calibration_for_parameter(self, parameter_dict, calibration):
        funcname = __name__ + '.set_calibration_for_parameter():'
        logger.debug(funcname)
        para = parameter_dict['parameter']
        para_parent = parameter_dict['parent']
        para_name = parameter_dict['name']
        para_index = parameter_dict['index']
        # Apply calibration to parameter (could also be a function)
        if para_parent.__class__.__base__ == pydantic.BaseModel:
            logger.debug(funcname + 'Setting parameter {} to calibration {}'.format(para_index, calibration))
            setattr(para_parent, para_index, calibration)
        elif isinstance(para_parent, list):
            logger.debug(funcname + 'Setting parameter in list {} to calibration {}'.format(para_index, calibration))
            para_parent[para_index] = calibration

    def logger_autocalibration(self):
        """
        Searches self.config.calibrations for calibration of loggers that can be autocalibrated,
        if loggers are found they are added to self.config.sensorconfigurations

        """
        funcname = __name__ + '.logger_autocalibration():'
        logger.debug(funcname)
        # Find calibrations for loggers
        loggers_autocal = ['DHFS50']
        # sn of the logger and the calibrations are the same for autocalibration
        for calibration in self.custom_config.calibrations:
            sn = calibration.sn
            if True:
                try:
                    sensor_model = calibration.sensor_model
                except:
                    sensor_model = None

                # Check if we can automatically find calibrations for the logger
                if sensor_model in loggers_autocal:
                    #print('Got calibration coefficients',sn)
                    latest = self.get_latest_callibration_coeffs(sn)
                    #print('Got calibration coefficients')
                    #print('latest',latest)

                    if latest is not None:
                        ret = self.add_sensorconfig(sn, sensor_model)
                        #print('Add loggerconfig', ret)
                        #print('latest',latest,latest.keys())
                        for parameter in latest.keys():
                            logger.debug(funcname + ' updating {:s}'.format(parameter))
                            self.custom_config.sensorconfigurations[sn]['parameter'][parameter] = latest[parameter]

                        if ret == True:
                            #print('Newlogger')
                            status = {}
                            self.newlogger_signal.emit(status)
                        else:
                            status = {}
                            self.newloggerconfig_signal.emit(status)
    def add_calibration_file(self, calfile, reload=True):
        funcname = __name__ + 'add_calibration_file():'
        logger.debug(funcname)
        calfiles_tmp = list(self.custom_config.calibration_files)
        if (calfile in calfiles_tmp) and reload:
            #print('Removing file first')
            self.rem_calibration_file(calfile)

        calfiles_tmp = list(self.custom_config.calibration_files)
        #print('Hallo',calfiles_tmp)
        if calfile not in calfiles_tmp:
            #print('Adding file 2',calfile)
            self.custom_config.calibration_files.append(calfile)
        else:
            logger.debug(funcname + ' File is already listed')

        self.populate_calibration()

    def rem_calibration_file(self, calfile):
        funcname = __name__ + 'rem_calibration_file():'
        logger.debug(funcname)
        calfiles_tmp = list(self.custom_config.calibration_files)
        if calfile in calfiles_tmp:
            calibration = self.read_calibration_file(calfile)
            calibration_json = json.dumps(calibration.model_dump())
            for cal_old in self.custom_config.calibrations:
                # Test if the calibration is existing
                if calibration_json == json.dumps(cal_old.model_dump()):
                    logger.debug(funcname + ' Removing calibration')
                    self.custom_config.calibration_files.remove(calfile)
                    self.custom_config.calibrations.remove(cal_old)
                    self.calfiles_processed.remove(calfile)
                    return True

    def populate_calibration(self):
        funcname = __name__ + '.populate_calibration()'
        logger.debug(funcname)
        self.calibration = {'sn': {}}
        fnames = []
        #if False:
        #    globstr = str(self.config['calibration_folder']) + '/*.yaml'
        #    for fname in glob.glob(globstr):
        #        fnames.append(fname)

        for fname in self.custom_config.calibration_files:
            fnames.append(str(fname))

        self.coeff_filenames = fnames

        for fname in fnames:
            if fname not in self.calfiles_processed:
                logger.debug(funcname + ' reading file {:s}'.format(fname))
                calibration = self.read_calibration_file(fname)
                if calibration is not None:
                    flag_add = self.add_calibration(calibration)
                    if flag_add:
                        self.calfiles_processed.append(fname)
            else:
                logger.debug(funcname + ' file {:s} already processed'.format(fname))


        #print(self.calibration['sn'].keys())
        self.logger_autocalibration()

    def add_calibration(self, calibration):
        """
        Adds a calibration to the calibration list, checks before, if the calibration exists
        calibration: calibration model
        """
        flag_new_calibration = True
        calibration_json = json.dumps(calibration.model_dump())
        for cal_old in self.custom_config.calibrations:
            if calibration_json == json.dumps(cal_old.model_dump()):
                flag_new_calibration = False
                break

        if flag_new_calibration:
            print('Sending new calibration signal')
            self.custom_config.calibrations.append(calibration)
            self.newcalibration_signal.emit()
            return True
        else:
            logger.warning('Calibration exists already')
            return False

    def read_calibration_file(self, fname):
        """
        Open and reads a calibration file, it will as well determine the type of calibration and call the proper function

        """
        funcname = __name__ + '.read_calibration_file():'
        logger.debug(funcname + 'Opening file {:s}'.format(fname))
        try:
            f = open(fname)
            data = yaml.safe_load(f)
            # print('data',data)
            if 'structure_version' in data.keys(): # Calibration model
                logger.debug(funcname + ' Version {} pydantic calibration model dump'.format(data['structure_version']))
                calmodels = [calibration_HF, calibration_NTC]
                for calmodel in calmodels:
                    try:
                        calibration = calmodel.model_validate(data)
                        print('Calibration', calibration)
                        return calibration
                    except:
                        pass

            elif 'sn' in data.keys():
                calibration = {}
                calibration['sn'] = {}
                logger.debug(funcname + ' Version 2 file')
                for sn in data['sn'].keys():  # Add the filename the data was loaded from
                    try:
                        calibration['sn'][sn]
                    except:
                        calibration['sn'][sn] = {}
                        #calibration['sn'][sn]['coeff_files'] = []

                    #calibration['sn'][sn]['coeff_files'].append(fname)
                    calibration['sn'][sn].update(data['sn'][sn])
                    calibration['sn'][sn]['original_file'] = fname
                #    data['sn'][sn].filenames = []
                #    data['sn'][sn].filenames.append(fname)
                # self.calibration['sn'].update(data['sn'])
                return calibration

            elif 'calibration_HF_SI' in data.keys(): # Version 1 file
                logger.debug(funcname + ' Version 1 file')

                sn = data['manufacturer_sn']
                if len(sn) ==0:
                    logger.debug(funcname + ' No serial number')
                else:
                    calibration = calibration_HF(sn=sn,coeff=data['calibration_HF_SI'],date=data['calibration_date'],sensor_model = data['series'])
                    #sn:
                    #    "9299":
                    #    model: F - 005 - 4
                    #    coeffs:
                    #        HF:
                    #        coeff: 8.98
                    #        parameter: HF
                    #        unit: W m-2
                    #        unitraw: mV
                    #        date: 1970 - 01 - 01_00 - 00 - 00
                    return calibration
                    if False:
                        calibration['sn'][sn] = {}
                        calibration['sn'][sn]['model'] = data['series']
                        calibration['sn'][sn]['original_file'] = fname
                        calibration['sn'][sn]['sn'] = sn
                        calibration['sn'][sn]['parameter'] = {}
                        calibration['sn'][sn]['parameter']['HF'] = {}
                        calibration['sn'][sn]['parameter']['HF']['coeff'] = data['calibration_HF_SI']
                        calibration['sn'][sn]['parameter']['HF']['unit'] = 'W m-2'
                        calibration['sn'][sn]['parameter']['HF']['unitraw'] = 'mW'
                        calibration['sn'][sn]['parameter']['HF']['date'] = data['calibration_date']

        except Exception as e:
            logger.exception(e)
            return None




    def get_latest_callibration_coeffs(self, sn):
        """
        Function is searching for the latest calibration
        """
        funcname = __name__ + '.get_latest_callibration_coeffs():'
        try:
            self.custom_config.calibrations[sn]
        except Exception as e:
            logger.warning('sn not found')
            return None

        parameters = {}
        parameters_latest = {}
        # Find all calibrations for sn
        calibrations_sn = []
        for calibration in self.custom_config.calibrations:
            if calibration.sn == sn:
                calibrations_sn.append(calibration)

        for calibration in calibrations_sn:
            if True:
                parameter = calibration.parameter
                try:
                    parameters[parameter]
                except:
                    parameters[parameter] = None
                    parameters_latest[parameter] = datetime.datetime(1,1,1,0,0,0)

                td = get_date_from_calibration(calibration,parameter)
                coeff = calibration['parameter'][parameter]
                #print('Got date',td)
                # Get the latest one here
                if parameters_latest[parameter] < td:
                    logger.debug(funcname + 'Found more recent coefficient for parameter {:s}'.format(parameter))
                    parameters[parameter] = coeff
                    parameters_latest[parameter] = td

        # Setting the parameters
        logger.debug(funcname + ' Setting the coefficient')

        #for parameter in parameters:
        #    self.config['calibrations'][sn]['used'][parameter] = parameters[parameter]

        return parameters

    def read_calibration_v02(self):
        """
        Reads calibration files of version 2. Written by the heatflow_calibration redvypr device
        """
        pass
    def read_calibration_v01(self):
        """
        Reads calibration files of version 1. Written by the calibration program
        """
        pass

    def thread_start(self):
        config = copy.deepcopy(self.custom_config)
        #config['calibration'] = self.calibration
        self.statusqueue_thread = threading.Thread(target=self.read_statusqueue, daemon=True)
        self.statusqueue_thread.start()
        super(Device, self).thread_start(config=config)


    def add_sensorconfig(self, sn, sensor_type, config = None, address=None):
        """

        """
        funcname = __name__ + '.add_sensorconfig():'
        logger.debug(funcname)
        try:
            sensor_new = sn not in self.custom_config.sensorconfigurations.keys()
            if sensor_new:
                logger.debug(funcname + 'Adding new sensor of type {:s} with sn {:s}'.format(sensor_type, sn))
                self.sensordata[sn] = {'__np__':0,'__raddr__':None,'__devices__':[]}
                self.sensordata_raw[sn] = []
                if sensor_type.lower() == 'hfv4ch':
                    sensor_config = logger_HFV4CH()
                    sensor_config.sn = sn
                    self.custom_config.sensorconfigurations[sn] = sensor_config
                    ret = True
                elif sensor_type.lower() == 'dhfs50':
                    sensor_config = sensor_DHFS50()
                    sensor_config.sn = sn
                    self.custom_config.sensorconfigurations[sn] = sensor_config
                    ret = True
                elif sensor_type.lower() == 'tar':
                    # Add a temperature array logger
                    sensor_config = sensor_TAR()
                    sensor_config.init_from_data(config)
                    self.custom_config.sensorconfigurations[sn] = sensor_config
                    ret = True
                else:
                    ret = None

                if ret is not None:
                    config = self.custom_config.model_dump()
                    print('config for new logger', config)
                    self.thread_command('config', data={'config': config})
                    status = {'sn':sn, 'sensor_type':sensor_type,'config':config}
                    self.newlogger_signal.emit(status)
            else:
                logger.debug(funcname + ' Configuration for {:s} is already existing'.format(sn))
                return False
        except:
            logger.warning(funcname + 'Could not add new sensor:',exc_info=True)




    def read_statusqueue(self):
        """
        Functions reads statusqueue
        """
        funcname = __name__ + '.read_statusqueue():'
        logger.debug(funcname)
        while True:
            status = self.statusqueue.get()
            logger.debug('Got status {:s}'.format(str(status)))
            if status == 'stop':
                #print('stopping')
                logger.debug(funcname + ' Stopping')
                return
            elif type(status) == dict:
                if status['status'] == 'newlogger':
                    logger.debug('New Sensor')
                    newsensor_address = RedvyprAddress(status['data'])
                    if status['packettype'] == 'HFV':
                        logger.info('New 4Channel logger')
                        sn = status['sn']
                        ret = self.add_sensorconfig(sn, sensor_type ='hfv4ch', address=newsensor_address)
                    elif status['packettype'] == 'TAR':
                        logger.info('New Temperature array')
                        sn = status['sn']
                        dataline = status['dataline']
                        ret = self.add_sensorconfig(sn, sensor_type='tar', config=dataline, address=newsensor_address)
                    elif status['packettype'] == 'HF':
                        logger.info('New DHFS50')
                        sn = status['sn']
                        ret = self.add_sensorconfig(sn, sensor_type ='dhfs50')
                    else:
                        logger.warning('Unknown packet type')

def find_calibration_for_logger(sn, calibration, caltype = 'latest'):
    """

    """
    pass

#
#
# Init Widget
#
#
class initDeviceWidget(QtWidgets.QWidget):
    connect = QtCore.pyqtSignal(
        RedvyprDevice)  # Signal requesting a connect of the datainqueue with available dataoutqueues of other devices

    def __init__(self, device = None):
        """
        Standard deviceinitwidget if the device is not providing one by itself.

        Args:
            device:
        """
        funcname = __name__ + '.__init__():'
        logger.debug(funcname)
        super().__init__()
        self.layout = QtWidgets.QGridLayout(self)
        self.config_widgets = []
        self.device = device
        self.create_sensorcoefficientWidget()
        self.config_widgets.append(self.sensorCoeffWidget)



        self.coeffbutton = QtWidgets.QPushButton('Calibration coefficients')
        self.coeffbutton.clicked.connect(self.show_coeffwidget)
        self.config_widgets.append(self.coeffbutton)

        self.__create_configLoggers_widget__()
        self.addLoggerButton = QtWidgets.QPushButton('Add sensor manually')
        self.addLoggerButton.clicked.connect(self.__configlogger_add_clicked__)
        self.config_widgets.append(self.addLoggerButton)

        self.coefffiles = QtWidgets.QListWidget()
        self.populate_coefffiles()
        self.create_sensorcoefffilesWidget()

        # Start-button
        self.startbutton = QtWidgets.QPushButton('Start')
        self.startbutton.clicked.connect(self.start_clicked)
        self.startbutton.setCheckable(True)
        # Process kill button (if thread)
        if (self.device.mp == 'multiprocess'):
            # Killbutton
            self.killbutton = QtWidgets.QPushButton('Kill process')
            self.killbutton.clicked.connect(self.kill_clicked)

        # Connect button
        self.conbutton = QtWidgets.QPushButton("Subscribe")
        self.conbutton.clicked.connect(self.connect_clicked)
        self.config_widgets.append(self.conbutton)
        sensorlabel = QtWidgets.QLabel('Sensors')
        sensorlabel.setAlignment(QtCore.Qt.AlignCenter)
        sensorlabel.setStyleSheet(''' font-size: 24px; font: bold''')
        self.layout.addWidget(sensorlabel, 0, 0)
        self.layout.addWidget(self.configLoggerWidgets['loggerlist'], 1, 0, 1, 4)
        self.layout.addWidget(self.coeffbutton, 2, 0, 1, 2)
        self.layout.addWidget(self.addLoggerButton,2, 2, 1, 2)
        self.layout.addWidget(self.conbutton, 3, 0, 1, 4)
        if (self.device.mp == 'multiprocess'):
            self.layout.addWidget(self.startbutton, 4, 0, 1, 3)
            self.layout.addWidget(self.killbutton, 4, 3)
        else:
            self.layout.addWidget(self.startbutton, 4, 0, 1, 4)

        # If the config is changed, update the device widget
        self.layout.setRowStretch(0, 0)
        self.layout.setRowStretch(1, 20)
        self.layout.setRowStretch(2, 1)
        self.layout.setRowStretch(3, 2)
        #GL.setColumnStretch(GL.columnCount(), 1)

        self.statustimer = QtCore.QTimer()
        self.statustimer.timeout.connect(self.update_buttons)
        self.statustimer.start(500)
        #self.config_widget.config_changed_flag.connect(self.config_changed)
        self.device.newlogger_signal.connect(self.__newlogger__)
        self.device.newloggerconfig_signal.connect(self.loggerconfig_changed)

        #config = XYplotWidget.configXYplot(title='Test')
        #self.plot_widget_HF_raw = XYplotWidget.XYplot(config=config, redvypr_device=self.device.redvypr)
        #self.plot_widget_HF_raw.show()

    def loggerconfig_changed(self):
        funcname = __name__ + '.loggerconfig_changed():'
        logger.debug(funcname)
        self.__update_configSensorList__()
    def __newlogger__(self,newlogger):
        funcname = __name__ + '.__newlogger__():'
        logger.debug(funcname)
        self.__update_configSensorList__()
    def finalize_init(self):
        funcname = __name__ + '.__create_configLoggers_widget__():'
        logger.debug(funcname)
        self.__update_configSensorList__()

    def __create_configLoggers_widget__(self):
        """

        """
        funcname = __name__ + '.__create_configLoggers_widget__():'
        logger.debug(funcname)

        self.configLoggerWidgets = {}
        #self.configLoggerWidgets['loggerlist'] = QtWidgets.QListWidget()  # A list with all loggers
        self.configLoggerWidgets['loggerlist'] = QtWidgets.QTableWidget()  # A list with all loggers
        #self.configLoggerWidgets['loggerlist'].currentRowChanged.connect(self.__configloggerlist_changed__)
        self.configLoggerWidgets['loggerconfig'] = QtWidgets.QWidget()
        self.configLoggerWidgets['loggerconfig_layout'] = QtWidgets.QVBoxLayout(self.configLoggerWidgets['loggerconfig'])
        self.__update_configSensorList__()

    def __configlogger_clicked__(self):
        funcname = __name__ + '__configlogger_clicked__():'
        logger.debug(funcname)
        item = self.sender().item
        layout = self.configLoggerWidgets['loggerconfig_layout']
        if item is not None:
            #print('item', item)
            #print('item', item.sn)
            #print('item', item.configwidget)
            while layout.count():
                itemold = layout.takeAt(0)
                widget = itemold.widget()
                widget.hide()

            self.configLoggerWidgets['loggerconfig_layout'].addWidget(item.configwidget)
            item.configwidget.show()
            self.configLoggerWidgets['loggerconfig'].show()


    def __configlogger_add_clicked__(self):
        self.configLoggerWidgets['addwidget'] = QtWidgets.QWidget()
        self.configLoggerWidgets['addwidget_layout'] = QtWidgets.QGridLayout(self.configLoggerWidgets['addwidget'])
        layout = self.configLoggerWidgets['addwidget_layout']
        self.configLoggerWidgets['loggercombo'] = QtWidgets.QComboBox()
        #print('Options',typing.get_type_hints(self.device.config.sensorconfigurations))
        print('Options',typing.get_type_hints(DeviceCustomConfig))
        sensorconfigurations_options_tmp = typing.get_type_hints(DeviceCustomConfig)['sensorconfigurations'] # This is a typing.Dict
        sensorconfigurations_options_tmp2 = typing.get_args(sensorconfigurations_options_tmp)[1] # This is a Union
        sensorconfigurations_options = typing.get_args(sensorconfigurations_options_tmp2) # This is a tuple with the arguments
        for hf_logger in sensorconfigurations_options:
            hf_logger_called = hf_logger()
            print('hf_logger',hf_logger)
            logger_model = hf_logger_called.logger_model
            logger_data = hf_logger # Store the hf_logger template
            #item = QtWidgets.QStandardItem(logger_model)
            #item.logger_model = logger_model
            self.configLoggerWidgets['loggercombo'].addItem(logger_model)
            cnt = self.configLoggerWidgets['loggercombo'].count()
            self.configLoggerWidgets['loggercombo'].setItemData(cnt-1,logger_data)
        self.configLoggerWidgets['loggeraddbutton'] = QtWidgets.QPushButton('Apply')
        self.configLoggerWidgets['loggeraddbutton'].clicked.connect(self.__configlogger_add_apply__)
        self.configLoggerWidgets['loggername'] = QtWidgets.QLineEdit()
        self.configLoggerWidgets['loggernamelabel'] = QtWidgets.QLabel('SN')
        self.configLoggerWidgets['loggertypelabel'] = QtWidgets.QLabel('Type')

        layout.addWidget(self.configLoggerWidgets['loggertypelabel'], 0, 0)
        layout.addWidget(self.configLoggerWidgets['loggernamelabel'], 0, 1)
        layout.addWidget(self.configLoggerWidgets['loggercombo'],1, 0)
        layout.addWidget(self.configLoggerWidgets['loggername'], 1, 1)
        layout.addWidget(self.configLoggerWidgets['loggeraddbutton'], 2, 0, 1, 2)
        self.configLoggerWidgets['addwidget'].show()

    def __configlogger_add_apply__(self):
        funcname = __name__ + '__configloggerlist_add__():'
        logger.debug(funcname)
        sn = str(self.configLoggerWidgets['loggername'].text())
        comboindex = self.configLoggerWidgets['loggercombo'].currentIndex()
        logger_template = self.configLoggerWidgets['loggercombo'].currentData()
        print('data',logger_template)
        loggertype = self.configLoggerWidgets['loggercombo'].currentText()
        if len(sn) > 0:
            logger.debug(funcname + ' Will add logger {:s} of type {:s}'.format(sn, loggertype))
        else:
            logger.debug(funcname + ' Specify a valid name')



        if sn in list(self.device.custom_config.sensorconfigurations.keys()):
            logger.warning(' Cannot add logger {:s}, as it exsits already'.format(sn))
            return None
        else:
            logger_new = logger_template()
            logger_new.sn = sn
            # Get the template
            logger_new.sn = sn
            self.device.custom_config.sensorconfigurations[sn] = logger_new

            self.__update_configSensorList__()
            self.configLoggerWidgets['addwidget'].close()

    def __update_configLoggerList_status__(self):
        """
        updates the status of the configloggerlist
        """
        funcname = __name__ + '____update_configLoggerList_status__():'
        logger.debug(funcname)

        colpub = 3
        coldev = 4
        colnp = 5
        for irow, sn in enumerate(self.device.custom_config.sensorconfigurations.keys()):
            try:
                raddr = self.device.sensordata[sn]['__raddr__']
                print('Raddr',raddr)
                item_publisher = QtWidgets.QTableWidgetItem(raddr.publisher)
                item_publisher_old = self.configLoggerWidgets['loggerlist'].item(irow, colpub)
                if item_publisher_old.text() != item_publisher.text():
                    self.configLoggerWidgets['loggerlist'].setItem(irow, colpub, item_publisher)

                devices = str(self.device.sensordata[sn]['__devices__'])
                item_devices = QtWidgets.QTableWidgetItem(devices)
                item_devices_old = self.configLoggerWidgets['loggerlist'].item(irow, coldev)
                if item_devices_old.text() != item_devices.text():
                    self.configLoggerWidgets['loggerlist'].setItem(irow, coldev, item_devices)

                np = self.device.sensordata[sn]['__np__']
                item_np = QtWidgets.QTableWidgetItem(str(np))
                item_np_old = self.configLoggerWidgets['loggerlist'].item(irow, colnp)
                if item_np_old.text() != item_np.text():
                    self.configLoggerWidgets['loggerlist'].setItem(irow, colnp, item_np)
            except:
                logger.info(funcname, exc_info=True)
                raddr = None

        self.configLoggerWidgets['loggerlist'].resizeColumnsToContents()

    def __update_configSensorList__(self):
        """
        update the table of sensor configuration widgets
        """
        funcname = __name__ + '____update_configLoggerList__():'
        logger.debug(funcname)
        self.configLoggerWidgets['loggerlist'].clear()
        nrows = len(self.device.custom_config.sensorconfigurations.keys())
        ncols = 6
        self.configLoggerWidgets['loggerlist'].setRowCount(nrows)
        self.configLoggerWidgets['loggerlist'].setColumnCount(ncols)
        self.configLoggerWidgets['loggerlist'].setHorizontalHeaderLabels(['SN','Sensortype','Configure','Publisher','Devices','Numpackets'])
        for irow,sn in enumerate(self.device.custom_config.sensorconfigurations.keys()):
            # Set blank everywhere
            for icol in range(ncols):
                item_blank = QtWidgets.QTableWidgetItem('')
                self.configLoggerWidgets['loggerlist'].setItem(irow, icol, item_blank)


            item = QtWidgets.QTableWidgetItem(str(sn))
            sensor_type = str(self.device.custom_config.sensorconfigurations[sn].sensor_type)
            #print('logger_short', logger_short)
            # Create a configuration widget for the sensor
            configwidget = self.__create_sensor_config_widget__(str(sn), sensor_type)
            item.configwidget = configwidget
            item.sn = sn
            #print('Item',item.configwidget)
            self.configLoggerWidgets['loggerlist'].setItem(irow,0,item)
            item_type = QtWidgets.QTableWidgetItem(sensor_type)
            self.configLoggerWidgets['loggerlist'].setItem(irow, 1, item_type)
            button_configure = QtWidgets.QPushButton('Configure')
            button_configure.clicked.connect(self.__configlogger_clicked__)
            button_configure.item = item
            self.configLoggerWidgets['loggerlist'].setCellWidget(irow, 2, button_configure)
            try:
                raddr = self.device.sensordata[sn]['__raddr__']
                #print('Raddr 2',raddr)
                item_publisher = QtWidgets.QTableWidgetItem(raddr.publisher)
                self.configLoggerWidgets['loggerlist'].setItem(irow, 4, item_publisher)
                #print('Done')
            except:
                logger.debug('invalid redvypr address',exc_info=True)
                raddr = None

        self.configLoggerWidgets['loggerlist'].resizeColumnsToContents()

    def __create_sensor_config_widget__(self, sn, sensor_type):
        funcname = __name__ + '__create_sensor_config_widget__()'
        logger.debug(funcname + ' sn: {:s}, sensor_short {:s}'.format(sn,sensor_type))
        cwidget = QtWidgets.QWidget()
        clayout = QtWidgets.QVBoxLayout(cwidget)
        clabel = QtWidgets.QLabel('Config of \n' + sn)
        clayout.addWidget(clabel)

        if sensor_type.lower() == 'generic logger':
            print('Generic logger ...')
        elif sensor_type.lower() == 'tar':
            logger.debug('Config widget for temperature array (TAR)')
            sensor = self.device.custom_config.sensorconfigurations[sn]
            #config_widget = TARWidget_config(self, sn=sn, redvypr_device=self.device)
            config_widget = sensorConfigWidget(sensor=sensor, calibrations=self.device.custom_config.calibrations, calibration_models=self.device.calibration_models, redvypr_device=self.device)
            clayout.addWidget(config_widget)
        elif sensor_type.lower() == 'dhfs50':
            logger.debug('DHFS50')
            sensor = self.device.custom_config.sensorconfigurations[sn]
            #config_widget = DHFS50Widget_config(sn = sn, redvypr_device = self.device)
            config_widget = sensorConfigWidget(sensor = sensor, calibrations = self.device.custom_config.calibrations, calibration_models=self.device.calibration_models, redvypr_device=self.device)
            clayout.addWidget(config_widget)
        elif sensor_type.lower() == 'hfv4ch':
            logger.debug('Four channel logger')
            config_widget = HFVWidget_config(sn = sn, redvypr_device = self.device)
            clayout.addWidget(config_widget)

        return cwidget

    def show_coefffileswidget(self):
        self.coefffileswidget.show()

    def create_sensorcoefffilesWidget(self):
        self.coefffileswidget = QtWidgets.QWidget()
        self.coefffileswidget_layout = QtWidgets.QHBoxLayout(self.coefffileswidget)
        self.coefffileswidget_layout.addWidget(self.coefffiles)

    def populate_coefffiles(self):
        for fname in self.device.coeff_filenames:
            self.coefffiles.addItem(fname)
    def show_coeffwidget(self):
        self.create_sensorcoefficientWidget()
        self.sensorCoeffWidget.show()




    def create_sensorcoefficientWidget(self):
        # Get the calibration models from the type hints of the config
        self.sensorCoeffWidget = sensorCoeffWidget(calibrations=self.device.custom_config.calibrations, redvypr_device=self.device, calibration_models = self.device.calibration_models)

    def config_changed(self):
        """
        """
        funcname = __name__ + '.config_changed():'
        logger.debug(funcname)

    def kill_clicked(self):
        button = self.sender()
        logger.debug("Kill device {:s}".format(self.device.name))
        self.device.kill_process()

    def start_clicked(self):
        button = self.sender()
        if button.isChecked():
            logger.debug("button pressed")
            button.setText('Starting')
            self.device.thread_start()
            # self.device_start.emit(self.device)
        else:
            logger.debug('button released')
            # button.setText('Stopping')
            self.startbutton.setChecked(True)
            self.device.thread_stop()

    def update_buttons(self):
        """ Updating all buttons depending on the thread status (if its alive, graying out things)
        """

        status = self.device.get_thread_status()
        thread_status = status['thread_running']
        # Running
        if (thread_status):
            self.startbutton.setText('Stop')
            self.startbutton.setChecked(True)
            for w in self.config_widgets:
                w.setEnabled(False)
        # Not running
        else:
            self.startbutton.setText('Start')
            for w in self.config_widgets:
                w.setEnabled(True)

            # Check if an error occured and the startbutton
            if (self.startbutton.isChecked()):
                self.startbutton.setChecked(False)
            # self.conbtn.setEnabled(True)

    def connect_clicked(self):
        button = self.sender()
        self.connect.emit(self.device)





class displayDeviceWidget(QtWidgets.QWidget):
    def __init__(self,device=None,tabwidget=None, deviceinitwidget=None):
        super(QtWidgets.QWidget, self).__init__()
        layout        = QtWidgets.QFormLayout(self)
        self.device   = device
        self.maxlen = 50000

        # Table for the raw data
        self.rawdatatableWidget = QtWidgets.QTableWidget(self)
        self.rawdatatableWidget.sensordata = {}
        self.rawrows    = []
        self.rawcolumns = []
        self.rawcolumns.append('') # Add an empty column for the Rowtype
        self.rawrows.append('Serialnum/MAC')
        self.rawrows.append('Model')

        layout.addRow(self.rawdatatableWidget)
        # Widget and Table for the converted data
        self.allconvertedWidget = QtWidgets.QWidget()
        self.allconvertedWidgetlayout = QtWidgets.QFormLayout(self.allconvertedWidget)
        self.converteddatableWidget = QtWidgets.QTableWidget(self.allconvertedWidget)
        self.converteddatableWidget.sensordata = {}
        self.convrows = []
        self.convcolumns = []
        self.convcolumns.append('')
        self.convrows.append('Serialnum/MAC')
        self.convrows.append('Model')

        self.allconvertedWidgetlayout.addRow(self.converteddatableWidget)


        # Sensorwidgets, sorted by sn, the widgets can be accessed by self.sensorwidget
        self.sensorwidget = QtWidgets.QTableWidget(self)
        self.create_sensorwidget()
        self.sensorwidgets = {}

        tabwidget.addTab(self,'Raw data')
        tabwidget.addTab(self.allconvertedWidget,'Converted data')
        tabwidget.addTab(self.sensorwidget, 'Sensors')
        # add already existing sensors
        for sn in self.device.custom_config.sensorconfigurations:
            sensor = self.device.custom_config.sensorconfigurations[sn]
            status = {}
            status['sn'] = sensor.sn
            status['sensor_type'] = sensor.sensor_type
            self.add_newsensor(status)
        # Connect a new sensor configuration signal
        self.device.newlogger_signal.connect(self.add_newsensor)


        #testwidget = DHFSWidget(sn='fdsfsdsf')
        #tabwidget.addTab(testwidget, 'Test')

        #sn = 'KAL1'
        #testwidget2 = HFVWidget(sn=sn,redvypr_device = self.device)
        #tabwidget.addTab(testwidget2, 'KAL1')
        #self.sensorwidgets[sn] = testwidget2

        #testwidget3 = HFVWidget(sn='KALTEST', redvypr_device=self.device)
        #tabwidget.addTab(testwidget3, 'KALTEST')

    def add_newsensor(self, status):
        """
        Function is called with when a new sensor signal is emitted by the device
        """
        funcname = __name__ + '.add_newsensor()'
        logger.debug(funcname)
        print('Status',status)
        sn = status['sn']
        sensor_type = status['sensor_type']
        try:
            self.sensorwidgets[sn]
        except:
            print('Will add new sensorwidget for sensor {}'.format(sn))
            self.add_sensorwidget(sn,sensor_type)


    def change_sensor(self, index):
        """
        Changes the active widget that is shown when clicked on the QListWidget
        """
        funcname = __name__ + '.change_sensor()'
        logger.debug(funcname)
        self.sensorstack.setCurrentIndex(index)

    def create_sensorwidget(self):
        funcname = __name__ + '.create_sensorwidget()'
        logger.debug(funcname)
        layout = QtWidgets.QVBoxLayout(self.sensorwidget)
        self.sensorlist = QtWidgets.QListWidget()
        self.sensorlist.currentRowChanged.connect(self.change_sensor)
        self.sensorstack = QtWidgets.QStackedWidget()
        splitter1 = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
        splitter1.addWidget(self.sensorlist)
        splitter1.addWidget(self.sensorstack)
        layout.addWidget(splitter1)


    def add_sensorwidget(self, sn, sensortype):
        funcname = __name__ + '.add_sensorwidget():'
        logger.debug(funcname+ ' Adding widget for {} and type {}'.format(sn, sensortype))
        # Add widgets
        # can beIMU, HFV, HF, TAR
        if sensortype.lower() == 'HFV'.lower():
            sensorwidget = HFVWidget(sn=sn, redvypr_device=self.device)
            self.sensorwidgets[sn] = sensorwidget
            self.sensorlist.addItem(sn)
            self.sensorstack.addWidget(sensorwidget)
        elif sensortype.lower() == 'TAR'.lower():
            sensorwidget = TARWidget(sn=sn, redvypr_device=self.device)
            self.sensorwidgets[sn] = sensorwidget
            self.sensorlist.addItem(sn)
            self.sensorstack.addWidget(sensorwidget)
        elif sensortype.lower() == 'HF'.lower():
            sensorwidget = DHFSWidget(sn=sn, redvypr_device=self.device)
            self.sensorwidgets[sn] = sensorwidget
            self.sensorlist.addItem(sn)
            self.sensorstack.addWidget(sensorwidget)
        else:# HF, IMU
            sensorwidget = QtWidgets.QLabel('Unknown Sensor {}'.format(sensortype))
            self.sensorwidgets[sn] = sensorwidget
            self.sensorlist.addItem(sn)
            self.sensorstack.addWidget(sensorwidget)

        # legacy
        #self.device.sensordata[sn] = {'__np__':0,'__raddr__':None,'__devices__':[]}
        #self.device.sensordata_raw[sn] = []



    def update_data(self, data):
        funcname = __name__ + '.update_data():'
        print(funcname)
        #return
        print(funcname + '  got data',data)
        try:
            sn = data['sn']
            # ptype is packettype
            ptype = data['type']
            datatype = data['datatype']
            # can be IMU, HFV, HF, TAR
            print('Ptype',ptype,sn,datatype)
            #self.device.config.sensorconfigurations[sn]
            #try:
            #    self.device.sensordata_raw[sn] # List for the raw datapackets
            #except:
            #    logger.debug('New sensor',exc_info=True)
            #    logger.debug(funcname + 'New sensor found with sn {:s}'.format(sn))
            #    try:
            #        self.add_sensorwidget(sn, sensortype=ptype)
            #    except:
            #        logger.debug(funcname, exc_info=True)
            #        logger.debug('Could not add sensorwidget for {:s}'.format(sn))
            #        return None

            print('Hallo',sn)
            raddr = RedvyprAddress(data)
            print('Hallo',raddr)
            # Add the data to self.device
            self.device.sensordata_raw[sn].append(data)
            self.device.sensordata[sn]['__np__'] += 1
            self.device.sensordata[sn]['__raddr__'] = raddr
            self.device.sensordata[sn]['__devices__'].append(raddr.devicename)
            self.device.sensordata[sn]['__devices__'] = list(set(self.device.sensordata[sn]['__devices__']))
            print('fdsfds')
            self.device.deviceinitwidget.__update_configLoggerList_status__()
            print('fdsfd_done')
            # Check if data needs to be removed
            if len(self.device.sensordata_raw[sn]) > self.maxlen:
                self.device.sensordata_raw[sn] = self.device.sensordata_raw[sn][-self.maxlen:]


            rdata = data_packets.Datapacket(data)
            parameter = rdata.datakeys(expand=1)
            #try:
            #    parameter.remove('type')  # Remove type
            #    parameter.remove('sn')  # Remove mac
            #except:
            #    logger.debug(exc_info=True)

            #print('Parameter',parameter)
            for p in parameter:
                # Update data
                try:
                    self.device.sensordata[sn][ptype]
                except:
                    self.device.sensordata[sn][ptype] = {}
                try:
                    self.device.sensordata[sn][ptype][p]
                except:
                    self.device.sensordata[sn][ptype][p] = []

                self.device.sensordata[sn][ptype][p].append(rdata[p])
                # Check if data needs to be removed
                if len(self.device.sensordata[sn][ptype][p]) > self.maxlen:
                    self.device.sensordata[sn][ptype][p] = self.device.sensordata[sn][ptype][p][-self.maxlen:]

            # Send the data to the sensorwidget
            print('Update sensor widget',sn)
            for w in self.sensorwidgets:
                try:
                   self.sensorwidgets[w].update_data(data)
                except:
                    logger.debug(funcname, exc_info=True)
            #try:
            #    self.sensorwidgets[sn].update(data)
            #except:
            #    logger.debug(funcname, exc_info=True)
            #    # logger.exception(e)

            # Update rows and columns
            nsensors  = len(list(self.device.sensordata_raw.keys()))

            if 'converted' in datatype.lower():
                #print('Updating converted table',ptype)
                rows = self.convrows
                columns = self.convcolumns
                table = self.converteddatableWidget
                # Store the last datapacket in each table
                table.sensordata[sn] = data
            else:
                print('Updating raw table',ptype)
                rows = self.rawrows
                columns = self.rawcolumns
                table = self.rawdatatableWidget
                print('rdata',rdata)
                table.sensordata[sn] = data

            flag_redraw = False
            for p in parameter:
                pstr = "{:s}.{:s}".format(data['type'],p)
                if pstr not in rows:
                    rows.append(pstr)
                    flag_redraw = True

            try:
                index = columns.index(sn)
            except:
                columns.append(sn)
                flag_redraw = True
                index = columns.index(sn)


            if flag_redraw:
                # Add publisher/device/numpackets to init
                # self.device.deviceinitwidget.
                #print('redraw table')
                #r = rows[0:2]
                #rows[0] = '0'
                #rows[1] = '1'
                #rows.sort()
                #rows[0] = r[0]
                #rows[1] = r[1]
                #print('Rows', rows)
                #table.clear()
                table.setRowCount(len(rows))
                table.setColumnCount(len(columns))
                table.setHorizontalHeaderLabels(columns)

                for irow,r in enumerate(rows):
                    if irow < 2:
                        r = str(irow)

                    item = QtWidgets.QTableWidgetItem(r)
                    table.setItem(irow, 0, item)

                table.sortItems(0)
                for irow, r in enumerate(rows):
                    if irow >= 2:
                        rows[irow] = table.item(irow,0).text()
                    else:
                        item = QtWidgets.QTableWidgetItem(rows[irow])
                        table.setItem(irow, 0, item)

                table.setVerticalHeaderLabels(rows)
                #mac_redraw = list(table.sensordata.keys())
                mac_redraw = [sn]
            else:
                mac_redraw = [sn]



            # Fill the table with data
            print('redraw',mac_redraw)
            for macdraw in mac_redraw:
                data_redraw = table.sensordata[macdraw]
                data_redraw_rdata = data_packets.Datapacket(data_redraw)
                parameter_redraw = data_redraw_rdata.datakeys(expand=1)
                #print('parameter redraw',parameter_redraw)
                index_col = columns.index(macdraw)
                #print('data redraw',data_)
                #print('filling table',macdraw, index_col)
                for p in parameter_redraw:
                    #print('P',p)
                    pstr = "{:s}.{:s}".format(data_redraw['type'], p)
                    index_row = rows.index(pstr)
                    # Packettype can be 'IMU','HF', 'HFSI'
                    #print('data redraw',data_redraw)
                    #print('fdsfdsfds',index_row,p,data_redraw_rdata[p])
                    if 'SI' in ptype:
                        if 'NTC' in p:
                            dstr = "{:.3f}".format(data_redraw_rdata[p])
                        elif 'HF' in p:
                            dstr = "{:.6f}".format(data_redraw_rdata[p])
                        elif p == 't': # Unix time stamp
                            dstr = "{:.3f}".format(data_redraw_rdata[p])
                        else:
                            dstr = str(data_redraw_rdata[p])
                    else:
                        dstr = str(data_redraw_rdata[p])

                    # Time string
                    if p == 't':
                        td = datetime.datetime.utcfromtimestamp(data_redraw[p])
                        dstr = td.strftime('%Y-%m-%d %H:%M:%S.%f')

                    item = QtWidgets.QTableWidgetItem(dstr)
                    #print('rows',rows)
                    #print('Item',p,pstr,index_row,index_col,dstr)
                    table.setItem(index_row,index_col,item)

            table.resizeColumnsToContents()
        except:
            logger.debug(funcname, exc_info=True)


        #print('Hallo',data)


